"""Provide encoding and decoding support for serialized workflow representations.

Reference https://docs.python.org/3/library/json.html#json.JSONDecoder and
https://docs.python.org/3/library/json.html#py-to-json-table describe the
trivial Python object conversions.
The core SCALE-MS encoder / decoder needs to manage the conversion of
additional types (scalems or otherwise, e.g. *bytes*) to/from these basic
Python types.

For JSON, we can provide an encoder for the *cls* parameter of json.dumps()
and we can provide a key-value pair processing dispatcher to the *object_pairs_hook*
parameter of json.loads()

We may also implement serialization schemes for document formats other than
JSON, such as the CWL schema.
"""
from __future__ import annotations

__all__ = ["BasicSerializable", "decode", "encode", "Shape"]

import abc
import collections.abc
import json
import logging
import os
import pathlib
import typing
import weakref

from scalems._types import BaseDecoded
from scalems._types import BaseEncodable
from scalems.exceptions import InternalError
from scalems.exceptions import MissingImplementationError
from scalems.exceptions import ProtocolError
from scalems.identifiers import EphemeralIdentifier
from scalems.identifiers import Identifier
from scalems.identifiers import TypeDataDescriptor
from scalems.identifiers import TypeIdentifier

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


class Shape(tuple):
    """Describe the data shape of a SCALEMS object."""

    def __new__(cls, elements: typing.Iterable):
        return super().__new__(cls, elements)

    def __init__(self, elements: typing.Iterable):
        """Initial implementation requires a sequence of integers.

        Software requirements include symbolic elements, TBD.
        """
        try:
            es = tuple(e for e in elements)
        except TypeError as e:
            raise e
        if len(es) < 1 or any(not isinstance(e, int) for e in es):
            raise TypeError("Shape is a sequence of 1 or more integers.")


# It could make sense to split the codec for native-Python encoding from the
# (de)serialization code in the future...


class SchemaDict(typing.TypedDict):
    """Schema for the member that labels an object's schema.

    This is just a type hint for the moment. The specification can be strengthened
    in the core data model and module constants provided for the schema comprising
    the full specification.

    Notes:
        * Python 3.9 provides a "frozenmap"
        * Consider a namedtuple, dataclass, or similar and make dict interconversion secondary.
        * We should clarify object model policies such as the invariance/covariance/contravariance
          of members through subtyping.

    TODO: Allow equality check
    TODO: Actually integrate with object support metaprogramming in the package.
    """

    spec: str
    name: str


class SymbolicDimensionSize(typing.TypedDict):
    DimensionSize: str


ShapeElement = typing.Union[int, SymbolicDimensionSize]


class FieldDict(typing.TypedDict):
    """Members of the *fields* member of a ResourceType."""

    schema: SchemaDict
    type: typing.List[str]
    shape: typing.List[ShapeElement]


FieldsType = typing.Mapping[str, FieldDict]


class TypeDict(typing.TypedDict):
    """Express the expected contents of a dictionary-based type description."""

    schema: SchemaDict
    implementation: typing.List[str]
    fields: FieldsType


class Encoded(typing.Protocol):
    """An Encoded object and all of its nested data are BaseEncodable.

    E.g.

        typing.Mapping[str, typing.Union[
            'Encoded',
            typing.Sequence['Encoded'], # should this be a union?
            str,
            int,
            float,
            bool,
            type(None)]]
    """

    # TODO: How should we implement this?


DispatchT = typing.TypeVar("DispatchT")
ResultT = typing.TypeVar("ResultT")
S = typing.TypeVar("S")


class Serializable(abc.ABC):
    """Base class for serialization behaviors.

    Subclassing takes care of registering encoders and decoders at
    module import.

    Serializable types must support encoding to and decoding from
    a small set of basic Python types for serialization schemes,
    such as JSON.
    """

    @abc.abstractmethod
    def encode(self) -> typing.Union[dict, list, tuple, str, int, float, bool, None]:
        ...

    @classmethod
    @abc.abstractmethod
    def decode(cls: typing.Type[S], *args, **kwargs) -> S:
        ...

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Create a weakref and add to Encoder._encoders with a finalizer to remove from same.
        # Similarly add to the Decoder dispatcher...


class PythonEncoder:
    """Encode SCALE-MS objects as basic Python data that is easily serialized.

    Extend the JSONEncoder for representations in the SCALE-MS data model by
    passing to the *default* argument of ``json.dumps()``,
    but note that it will only be used for objects that JSONEncoder does not already
    resolve.

    Note that json.dump and json.dumps only use the *default* call-back when the *cls* encoder does not
    have an implementation for an object type. To preempt standard processing by the JSONEncoder,
    you must provide a *cls* that overrides the encoding methods as documented at
    https://docs.python.org/3/library/json.html#json.JSONEncoder.encode to produce a string.
    This is _not_ what the *encode* method of this class does.

    Alternatively, encode object(s) first, and pass the resulting Python object to a regular call to json.dumps.
    """

    # Note that the following are equivalent.
    #     json.loads(s, *, cls=None, **kw)
    #     json.JSONDecoder(**kw).decode(s)
    # Note that the following are equivalent.
    #     json.dumps(obj, *, cls=None, **kw)
    #     json.JSONEncoder(**kw).encode(obj)

    # We use WeakKeyDictionary because the keys are likely to be classes,
    # and we don't intend to extend the life of the type objects (which might be temporary).
    _dispatchers: typing.ClassVar[
        typing.MutableMapping[type, typing.Callable[[DispatchT], BaseEncodable]]
    ] = weakref.WeakKeyDictionary()

    @classmethod
    def register(cls, dtype: typing.Type[DispatchT], handler: typing.Callable[[DispatchT], BaseEncodable]):
        # Note that we don't expect references to bound methods to extend the life of the type.
        # TODO: confirm this assumption in a unit test.
        if not isinstance(dtype, type):
            raise TypeError("We use `isinstance(obj, dtype)` for dispatching, so *dtype* must be a `type` object.")
        if dtype in cls._dispatchers:
            raise ProtocolError(f"Encodable type {dtype} appears to be registered already.")
        cls._dispatchers[dtype] = handler

    @classmethod
    def unregister(cls, dtype: typing.Type[DispatchT]):
        # As long as we use a WeakKeyDictionary, explicit unregistration should not be necessary.
        del cls._dispatchers[dtype]

    @classmethod
    def encode(cls, obj) -> BaseEncodable:
        """Convert an object of a registered type to a representation as a basic Python object."""
        # Currently, we iterate because we may be using abstract types for encoding.
        # If we find that we are using concrete types and/or we need more performance,
        # or if we just find that the list gets enormous, we can inspect the object first
        # to derive a dtype key that we can look up directly.
        # Warning: we should be careful not to let objects unexpectedly match multiple entries.
        for dtype, dispatch in cls._dispatchers.items():
            if isinstance(obj, typing.cast(type, dtype)):
                return dispatch(obj)
        raise TypeError(f"No registered dispatching for {repr(obj)}")

    def __call__(self, obj) -> BaseEncodable:
        return self.encode(obj)


class UnboundObject(typing.Protocol):
    """A prototypical instance of a workflow item not bound to a workflow.

    Generally, SCALEMS objects are items in a managed workflow.
    """

    def shape(self) -> Shape:
        ...

    def dtype(self) -> TypeIdentifier:
        ...

    def encode(self) -> BaseEncodable:
        ...


class PythonDecoder:
    """Convert dictionary representations to SCALE-MS objects for registered types.

    Dictionaries are recognized as SCALE-MS object representations with a minimal heuristic.

    If the object (dict) contains a *'schema'* key, and the value
    is a dict, the *'spec'* member of the dict is retrieved. If the *'spec'* member exists and
    names a recognized schema specification, the object is dispatched according to the schema
    specification.

    Otherwise, if the object contains a *'type'* key, identifying a recognizable registered type,
    the object is dispatched to the decoder registered for that type.

    For more information, refer to the :doc:`serialization` and :doc:`datamodel` documentation.

    .. todo:: Consider specifying a package metadata resource group to allow packages to register
              additional schema through an idiomatic plugin system.
    Refs:
     * https://packaging.python.org/guides/creating-and-discovering-plugins/
     * https://setuptools.readthedocs.io/en/latest/userguide/entry_point.html#dynamic-discovery-of-services-and-plugins
    """

    _dispatchers: typing.MutableMapping[TypeIdentifier, typing.Callable] = dict()

    # Depending on what the callables are, we may want a weakref.WeakValueDictionary() or we may not!

    @classmethod
    def register(cls, typeid: TypeIdentifier, handler: typing.Callable):
        # Normalize typeid
        typeid = TypeIdentifier.copy_from(typeid)
        if typeid in cls._dispatchers:
            raise ProtocolError("Type appears to be registered already.")
        cls._dispatchers[typeid] = handler

    @classmethod
    def unregister(cls, typeid: TypeIdentifier):
        del cls._dispatchers[typeid]

    @classmethod
    def get_decoder(cls, typeid) -> typing.Union[None, typing.Callable]:
        # Normalize the type identifier.
        try:
            identifier: typing.Optional[TypeIdentifier] = TypeIdentifier.copy_from(typeid)
            assert isinstance(identifier, TypeIdentifier)
            typename = identifier.name()
        except TypeError:
            try:
                typename = str(typeid)
            except TypeError:
                typename = repr(typeid)
            identifier = None
        # Use the (hashable) normalized form to look up a decoder for dispatching.
        if identifier is None or identifier not in cls._dispatchers:
            raise TypeError("No decoder registered for {}".format(typename))
        return cls._dispatchers[identifier]

    @classmethod
    def decode(cls, obj) -> typing.Union[UnboundObject, BaseDecoded]:  # noqa: C901
        """Create unbound SCALE-MS objects from their basic Python representations.

        We assume this is called in a bottom-up manner as a nested record is deserialized.

        Unrecognized objects are returned unaltered because they may be members
        of an enclosing object with appropriate dispatching.

        .. todo:: Consider where to register transcoders for compatible/virtual types.
                  E.g. Infer np.array(..., dtype=int) -> scalems.Integer
                  This is a small number of cases, since we can lean on the descriptors in the buffer protocol.
        """
        if not isinstance(obj, dict):
            # Probably don't have any special handling for such objects until we know what they are nested in.
            ...
        else:
            assert isinstance(obj, dict)
            if "schema" in obj:
                # We currently have very limited schema processing.
                try:
                    spec = obj["schema"]["spec"]
                except KeyError:
                    spec = None
                if not isinstance(spec, str) or spec != "scalems.v0":
                    # That's fine...
                    logger.info("Unrecognized *schema* when decoding object.")
                    return obj
                if "name" not in obj["schema"] or not isinstance(obj["schema"]["name"], str):
                    raise InternalError("Invalid schema.")
                else:
                    # schema = obj['schema']['name']
                    ...
                # Dispatch the object...
                ...
                raise MissingImplementationError(
                    "We do not yet support dynamic type registration through the work record."
                )

            if "type" in obj:
                # Dispatch the decoding according to the type.
                try:
                    dispatch = cls.get_decoder(obj["type"])
                except TypeError:
                    dispatch = BasicSerializable.decode
                if dispatch is not None:
                    return dispatch(obj)
        # Just return un-recognized objects unaltered.
        return obj

    def __call__(self, obj) -> typing.Union[UnboundObject, BaseDecoded]:
        return self.decode(obj)


encode = PythonEncoder()
decode = PythonDecoder()

# TODO: use stronger check for UID, or bytes-based objects.
encode.register(dtype=bytes, handler=bytes.hex)
encode.register(dtype=pathlib.Path, handler=os.fsdecode)
# TODO: Check that this dispatches correctly and update the type hinting.
# mypy gives "error: Only concrete class can be given where "Type[_PathLike[_AnyStr_co]]" is expected"
encode.register(dtype=os.PathLike, handler=os.fsdecode)


# Note that the low-level encoding/decoding is not necessarily symmetric because nested objects may be decoded
# according to the schema of a parent object.
# decode.register()


# A SCALE-MS "Serializable Type".
# TODO: use a Protocol or other constraint.
ST = typing.TypeVar("ST", bound="BasicSerializable")


class BasicSerializable(UnboundObject):
    __label: typing.Optional[str] = None
    __identity: Identifier
    _shape: Shape
    data: collections.abc.Container

    _data_encoder: typing.Callable
    _data_decoder: typing.Callable

    _dtype = TypeDataDescriptor(base_type=TypeIdentifier(("scalems", "BasicSerializable")))

    def dtype(self) -> TypeIdentifier:
        # Part of the decision of whether to use a property or a method
        # is whether we want to normalize on dtype as an instance or class characteristic.
        # Initially, we are using inheritance to get registration behavior through metaprogramming.
        # In other words, the real question may be how we want to handle registration.
        return self._dtype

    def __init__(self, data, *, dtype, shape=(1,), label=None, identity=None):
        if identity is None:
            # TODO: Calculate an appropriate identifier
            self.__identity = EphemeralIdentifier()
        else:
            # TODO: Validate identity
            self.__identity = identity
        self.__label = str(label)

        attrname = BasicSerializable._dtype.attr_name
        setattr(self, attrname, TypeIdentifier.copy_from(dtype))

        self._shape = Shape(shape)
        # TODO: validate data dtype and shape.
        # TODO: Ensure that we retain a reference to read-only data.
        # TODO: Allow a secondary localized / optimized / implementation-specific version of data.
        self.data = data

    def identity(self):
        return self.__identity

    def label(self):
        return str(self.__label)

    def shape(self):
        return Shape(self._shape)

    def encode(self) -> dict:
        representation = {
            "label": self.label(),
            "identity": str(self.identity()),
            "type": self.dtype().encode(),
            "shape": tuple(self.shape()),
            "data": self.data,  # TODO: use self._data_encoder()
        }
        return representation

    @classmethod
    def decode(cls: typing.Type[ST], encoded: dict) -> ST:
        if not isinstance(encoded, collections.abc.Mapping) or "type" not in encoded:
            raise TypeError("Expected a dictionary with a *type* specification for decoding.")
        dtype = TypeIdentifier.copy_from(encoded["type"])
        label = encoded.get("label", None)
        identity = encoded.get("identity")  # TODO: verify and use type schema to decode.
        shape = Shape(encoded["shape"])
        data = encoded["data"]  # TODO: use type schema / self._data_decoder to decode.
        logger.debug(f"Decoding {identity} as BasicSerializable.")
        return cls(label=label, identity=identity, dtype=dtype, shape=shape, data=data)

    def __init_subclass__(cls, **kwargs):
        assert cls is not BasicSerializable

        # Handle SCALE-MS Type registration.
        base = kwargs.pop("base_type", None)
        if base is not None:
            typeid = TypeIdentifier.copy_from(base)
        else:
            typeid = [str(cls.__module__)] + cls.__qualname__.split(".")
        registry = BasicSerializable._dtype.base
        if cls in registry and registry[cls] is not None:
            # This may be a customization or extension point in the future, but not today...
            raise ProtocolError("Subclassing BasicSerializable for a Type that is already registered.")
        BasicSerializable._dtype.base[cls] = typeid

        # Register encoder for all subclasses. Register the default encoder if not overridden.
        # Note: This does not allow us to retain the identity of *cls* for when we call the helpers.
        # We may require such information for encoder functions to know why they are being called.
        encoder = getattr(cls, "encode", BasicSerializable.encode)
        PythonEncoder.register(cls, encoder)

        # Optionally, register a new decoder.
        # If no decoder is provided, use the basic decoder.
        if hasattr(cls, "decode") and callable(cls.decode):
            _decoder = weakref.WeakMethod(cls.decode)

            # Note that we do not require that the decoded object is actually
            # an instance of cls.

            def _decode(encoded: dict):
                decoder = _decoder()
                if decoder is None:
                    raise ProtocolError("Decoding a type that has already been de-registered.")
                return decoder(encoded)

            PythonDecoder.register(cls._dtype, _decode)

        # TODO: Register optional instance initializer / input processor.
        # Allow instances to be created with something other than a single-argument
        # of the registered Input type.

        # TODO: Register/generate UI helper.
        # From the user's perspective, an importable module function interacts
        # with the WorkflowManager to add workflow items and return a handle.
        # Do we want to somehow generate an entry-point command

        # TODO: Register result dispatcher(s).
        # An AbstractDataSource must register a dispatcher to an implementation
        # that produces a ConcreteDataSource that provides the registered Result type.
        # A ConcreteDataSource must provide support for checksum calculation and verification.
        # Optionally, ConcreteDataSource may provide facilities to convert to/from
        # native Python objects or other types (such as .npz files).

        # Proceed dispatching along the MRO, per documented Python data model.
        super().__init_subclass__(**kwargs)


def compact_json(obj) -> str:
    """Produce the compact JSON string for the encodable object."""
    # Use the extensible Encoder from the serialization module, but apply some output formatting.
    string = json.dumps(obj, default=encode, ensure_ascii=True, separators=(",", ":"), sort_keys=True)
    return string


class JsonObjectPairsDispatcher:
    """Decode key/value pairs from JSON objects into SCALE-MS objects.

    Provides a place to register different type handlers.

    Each JSON *object* deserialized by the JSON Decoder is passed as a sequence
    of (key, value) pairs. The result is returned instead of the usual *dict(pairs)*.

    We don't have interaction with state or nesting, so we may have to examine the
    values for Python objects that have already been decoded to see if additional
    processing is necessary once the context of the key/value pair is known.
    """

    def __call__(self, key, value):
        ...


# def object_pair_decoder(context, object_pairs: typing.Iterable[typing.Tuple[str, typing.Any]])\
#         -> typing.Iterable[ItemView]:
#     """Decode named objects, updating the managed workflow as appropriate.
#
#     For object pairs representing complete workflow items, get a handle to a managed workflow item.
#     If the key is already managed, update the the managed item or raise an error if the managed item
#     is not consistent with the received item.
#
#     Note that responsibilities for validating work graphs, data flow, and compatibility are delegated to the
#     WorkflowManager and the registered data and command types. It does not make sense to call this function without
#     a proper WorkflowManager. For low-level testing or other use cases, consider directly using PythonDecoder.
#
#     To extend json.load() or json.loads(), use functools.partial to bind a workflow context, and pass the
#     partially bound function as the *object_pairs_hook* argument to the json deserializer.
#     """
#     # We would generally want to deserialize directly into a WorkflowManager. We could write this as a free function
#     # and optionally bind it as a method. We could also make it a singledispatch function or a singledispatchmethod.
#     # These are probably not mutually exclusive.
#     for key, value in object_pairs:
#         # dispatch decoding for value
#         # if is_workflowitem(decoded):
#         #    identity = validate_id(key, decoded)
#         #    record = {identity: decoded}
#         item_view = context.add_item(record)
#         yield item_view

Key = typing.Union[str, int, slice]


class OperationNode(abc.ABC):
    """Abstract interface for operation node references."""

    @abc.abstractmethod
    def to_json(self, **json_args) -> str:
        """Serialize the node to a JSON record."""
        ...

    @classmethod
    def from_json(cls, serialized: str):
        """Creation method to deserialize a JSON record."""
        # TODO: We could, of course, dispatch to registered subclasses,
        #  but this is deferred until refactoring converts node references into
        #  views into a Context.
        ...

    @abc.abstractmethod
    def fingerprint(self):
        """Get the unique identifying information for the node."""
        ...



"""Protocols and typing tools for identifying SCALE-MS objects and interfaces.

TODO: Publish some schema to scalems.org for easy reference and verifiable UUID hashing.
"""

from __future__ import annotations

__all__ = [
    "NAMESPACE_SCALEMS",
    "FingerprintHash",
    "Identifier",
    "NamedIdentifier",
    "ResourceIdentifier",
    "TypeIdentifier",
    "OperationIdentifier",
    "TypeDataDescriptor",
    "EphemeralIdentifier",
]

import abc
import builtins
import logging
import os
import typing
import uuid
import weakref

from scalems._types import BaseEncodable
from scalems.exceptions import InternalError
from scalems.exceptions import ProtocolError

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))

# TODO: Use a URL to the schema or specification.
NAMESPACE_SCALEMS: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_DNS, "scalems.org")

FingerprintHash = typing.NewType("FingerprintHash", bytes)
"""The fingerprint hash is a 32-byte sequence containing a SHA256 digest."""


@typing.runtime_checkable
class Identifier(typing.Hashable, typing.Protocol):
    """SCALE-MS object identifiers support this protocol.

    Identifiers may be implemented in terms of a hashing scheme, RFC 4122 UUID,
    or other encoding appropriate for the scope of claimed uniqueness and
    reusability (cacheable).

    Namespace UUIDs are appropriate for strongly specified names, such as operation
    implementation identifiers. The 48 bits of data are sufficient to identify graph
    nodes at session scope. At workflow scope, we need additional semantics about what
    should persist or not.

    Concrete data warrants a 128-bit or 256-bit checksum.

    Design notes:
        Several qualifications may be necessary regarding the use of an identifier,
        which may warrant annotation accessible through the Identifier instance.
        Additional proposed data members include:

        * scope: Scope in which the Identifier is effective and unique.
        * reproducible: Whether results will have the same identity if re-executed.
          Relates to behaviors in situations such as missing cache data.
        * concrete: Is this a concrete object or something more abstract?

    """

    concrete: bool
    """Is this a concrete object or something more abstract?"""

    @abc.abstractmethod
    def bytes(self) -> bytes:
        """A consistent bytes representation of the identity.

        The core interface provided by Identifiers.

        Note that the identity (and the value returned by self.bytes()) must be immutable
        for the life of the object.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __str__(self) -> str:
        """Represent the identifier in a form suitable for serialized records.

        All Identifier subclasses must explicitly define the string representation,
        but may ``return super().__str__(self)`` for a suitable default
        (a hexadecimal encoding of the core data).

        By default, the string representation is the basis for the stub used
        for filesystem objects. To change this, override the __fspath__() method.
        """
        return self.bytes().hex()

    def encode(self) -> BaseEncodable:
        """Get a canonical encoding of the identifier as a native Python object.

        This is the method that will be used to produce serialized workflow records.

        By default, the string representation (self.__str__()) is used.
        Subclasses may override, as long as suitable decoding is possible and provided.
        """
        return str(self)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other.bytes() == self.bytes()

    def __hash__(self) -> int:
        # Note that the result of the `hash()` built-in is truncated to
        # the size of a `Py_ssize_t`, but two objects that compare equal must
        # return the same value for `__hash__()`, so we return the full value.
        return self.__index__()

    def __fspath__(self) -> str:
        """Get a representation suitable for naming a filesystem object.

        Support the `os.PathLike` protocol for `os.fspath`.
        """
        path = os.fsencode(str(self))
        return str(path)

    def __bytes__(self) -> builtins.bytes:
        """Get the network ordered byte sequence for the raw identifier."""
        return self.bytes()

    def __index__(self) -> int:
        """Support integer conversions, including the `hex()` builtin function."""
        return int.from_bytes(self.bytes(), "big")


class EphemeralIdentifier(Identifier):
    """Process-scoped UUID based identifier.

    Not reproducible. Useful for tracking objects within a single process scope.
    """

    def __init__(self, node=None, clock_seq=None):
        self._data = uuid.uuid1(node, clock_seq)

    def bytes(self):
        return self._data.bytes

    def __str__(self) -> str:
        return str(self._data)


class NamedIdentifier(Identifier):
    """A name with strong identity semantics, represented with a UUID."""

    # TODO: facility to annotate scope
    # TODO: facility to annotate reproducibility
    # TODO: facility to indicate whether this is a reference to concrete data or not.
    def __init__(self, nested_name: typing.Sequence[str]):
        try:
            if isinstance(nested_name, (str, bytes)):
                raise TypeError("Wrong kind of iterable.")
            self._name_tuple = tuple(str(part) for part in nested_name)
        except TypeError as e:
            raise TypeError(f"Could not construct {self.__class__.__name__} from {repr(nested_name)}") from e
        else:
            self._data = uuid.uuid5(NAMESPACE_SCALEMS, ".".join(self._name_tuple))
        # TODO: The instance should track a context in which the uuid can be resolved.

    def bytes(self):
        return self._data.bytes

    def __str__(self) -> str:
        return str(self._data)

    def encode(self) -> BaseEncodable:
        return self._name_tuple


class ResourceIdentifier(Identifier):
    # TODO: facility to annotate scope
    # TODO: facility to annotate reproducibility
    # TODO: facility to indicate whether this is a reference to concrete data or not.
    def __init__(self, fingerprint: bytes):
        self._data = bytes(fingerprint)
        # Expecting a 256-bit SHA256 hash digest
        if len(self._data) != 32:
            raise InternalError(f"Expected a 256-bit hash digest. Got {repr(fingerprint)}")

    def bytes(self):
        return self._data

    def __str__(self) -> str:
        return self._data.hex()


class TypeIdentifier(NamedIdentifier):
    def name(self):
        return ".".join(self._name_tuple)

    def scoped_name(self) -> typing.Tuple[str, ...]:
        return self._name_tuple

    @classmethod
    def copy_from(cls, typeid) -> "TypeIdentifier":
        """Create a new TypeIdentifier instance describing the same type as the source.

        .. todo:: We need a generic way to determine the (registered) virtual type
                  of an object, but that doesn't belong here.
        """
        if isinstance(typeid, NamedIdentifier):
            # Copy from a compatible object.
            return cls(typeid._name_tuple)
        elif isinstance(typeid, (list, tuple)):
            # Create from the usual initialization parameter type.
            return cls(typeid)
        elif isinstance(typeid, type):
            # Try to generate an identifier based on a defined class.
            #
            # Consider disallowing TypeIdentifiers for non-importable types.
            # (Requires testing and enforcement.)
            # Consider allowing class objects to self-report their type.
            if typeid.__module__ is not None:
                fully_qualified_name = ".".join((typeid.__module__, typeid.__qualname__))
            else:
                fully_qualified_name = str(typeid.__qualname__)
            return cls.copy_from(fully_qualified_name)
        elif isinstance(typeid, str):
            # Conveniently try to convert string representations back into
            # the namespace sequence representation.
            # TODO: First check if the string is a UUID or other reference form
            #  for a registered type.
            return cls.copy_from(tuple(typeid.split(".")))
        # TODO: Is there a dictionary form that we should allow?
        else:
            raise TypeError(f"Cannot create a TypeIdentifier from {repr(typeid)}")


class TypeDataDescriptor:
    """Implement the *dtype* attribute.

    The TypeDataDescriptor object is instantiated to implement the
    BasicSerializable.base_type dynamic attribute.

    Attributes:
        name: Name of the attribute provided by the data descriptor.
        base: TypeIdentifier associated with the Python class.

    *name* can be provided at initialization, but is overridden during class
    definition when TypeDataDescriptor is used in the usual way (as a data descriptor
    instantiated during class definition).

    At least for now, *name* is required to be ``_dtype``.

    *attr_name* is derived from *name* at access time. For now, it is always
    ``__dtype``.

    Instances of the Python class may have their own *dtype*. For the SCALE-MS
    data model, TypeIdentifier is an instance attribute rather than a class attribute.
    If an instance did not set ``self.__dtype`` at initialization, the descriptor
    returns *base* for the instance's class.

    *base* is the (default) SCALEMS TypeIdentifier for the class using the descriptor.
    For a class using the data descriptor, *base* is inferred from the class
    __module__ and __qualname__ attributes, if not provided through the class definition.

    A single data descriptor instance is used for a class hierarchy to encapsulate
    the meta-programming for UnboundObject classes without invoking Python metaclass
    arcana (so far). At module import, a TypeDataDescriptor is instantiated for
    BasicSerializable._dtype. The data descriptor instance keeps a weakref.WeakKeyDict
    mapping type objects (classes) to the TypeDataDescriptor details for classes
    other than BasicSerializable. (BasicSerializable._dtype always produces
    ``TypeIdentifier(('scalems', 'BasicSerializable'))``.)
    The mapping is updated whenever BasicSerializable is subclassed.
    """

    base: typing.MutableMapping[type, TypeIdentifier]
    _original_owner_type: typing.Optional[TypeIdentifier]

    @property
    def attr_name(self):
        """Name of the instance data member used by this descriptor for storage."""
        return "_owner" + self.name

    def __init__(self, name: str = None, base_type: TypeIdentifier = None):
        # Note that the descriptor instance is not fully initialized until it is
        # further processed during the creation of the owning class.
        self.name = name
        if base_type is not None:
            self._original_owner_type = TypeIdentifier.copy_from(base_type)
        else:
            self._original_owner_type = None
        self.base = weakref.WeakKeyDictionary()

    def __set_name__(self, owner, name):
        # Called by type.__new__ during class creation to allow customization.
        # Let's start with strict naming requirements for early implementations,
        # and explicitly forbid multiple instances of this data descriptor implementation
        # in the same class.
        # Note that __set_name__ is only called at most once, by type.__new__
        # for a class definition in which the descriptor is instantiated.
        # In other words, __set_name__ is called for the base class, only, and
        # __init_subclass__ is called for derived classes, only.
        if name != "_dtype":
            raise ProtocolError("TypeDataDescriptor has a strict naming protocol. Only use for a `_dtype` attribute.")
        self.name = name
        if hasattr(owner, self.attr_name):
            raise ProtocolError(
                f"No storage for data descriptor. {repr(owner)} already has an attribute named {self.attr_name}."
            )

        assert owner not in self.base
        assert len(self.base) == 0
        logger.debug(f"Initializing base class {owner} ownership of TypeDataDescriptor.")
        self._original_owner = weakref.ref(owner)
        if self._original_owner_type is None:
            self._original_owner_type = TypeIdentifier.copy_from(
                [str(owner.__module__)] + owner.__qualname__.split(".")
            )
        self.base[owner] = TypeIdentifier.copy_from(self._original_owner_type)

    def __get__(self, instance, owner):
        # Note that instance==None when called through the *owner* (as a class attribute).
        if instance is None:
            if owner is self._original_owner():
                return self
            return self.base[owner]
        return getattr(instance, self.attr_name, self.base[owner])


class OperationIdentifier(tuple):
    """Python structure to identify an API Operation implementation.

    Operations are identified with a nested scope. The OperationIdentifier
    is a sequence of identifiers such that the operation_name() is the final
    element, and the preceding subsequence comprises the namespace().

    Conventional string representation of the entire identifier uses a period
    (``.``) delimiter.
    """

    def namespace(self):
        return tuple(self[0:-2])

    def operation_name(self):
        return self[-1]

    def __str__(self):
        return ".".join(self)



"""
Gromacs simulation tools.

Preparation and output manipulation use command line tools.
Simulation is executed with gmxapi.
"""
# Declare the public interface of this wrapper module.
__all__ = ["make_input", "modify_input", "simulate"]

##########################
# New plan:
# Implement basic gmxapi examples directly without generalism or type checking:
# Create simple dictionaries and add them to the workflow manager.
# 0. Write gromacs wrapper in terms of dicts. Processing is done via scalems entry points.
# 1. Hard-code handling in WorkflowManager.
# 2. Resume handling of dict and JSON representations in fingerprint and serialization
# modules.
# 3. Refine Python data model and build framework.
###########################
import functools
import logging

# import gmxapi
import os
import pathlib
import typing
from typing import Sequence

import scalems.workflow
from scalems.exceptions import MissingImplementationError
from scalems.unique import next_monotonic_integer as _next_int
from scalems.workflow import WorkflowManager

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


# TODO: REMOVE. Follow with fingerprint based UID ASAP.
def _next_uid() -> bytes:
    # Note that network byte order is equivalent to "big endian".
    return _next_int().to_bytes(32, "big")


def _executable(
    argv: Sequence[str],
    inputs: Sequence[str] = (),
    outputs: Sequence[str] = (),
    stdin: typing.Iterable[str] = (),
    resources: typing.Mapping[str, str] = None,
    environment: typing.Mapping[str, str] = None,
    stdout: str = "stdout.txt",
    stderr: str = "stderr.txt",
):
    if resources is None:
        resources = {}
    if environment is None:
        environment = {}

    input_node = {
        # Static properties.
        "object_type": "ExecutableInput",
        "uid": _next_uid().hex(),
        # Fields may have dependencies.
        "data": {"argv": list(argv), "inputs": list(inputs), "stdin": stdin, "environment": dict(**environment)},
    }
    # Results are self-referential until a concrete result is available.
    input_node["result"] = input_node["uid"]

    task_node: dict = {
        # Static properties.
        "object_type": "Executable",
        "uid": _next_uid().hex(),
        "data": {
            "stdout": "stdout.txt",
            "stderr": "stderr.txt",
        },
        # Fields may have dependencies.
        "input": {
            "ExecutableInput": input_node["uid"],
            "resources": dict(**resources),
            # The files produced by a program may depend on the inputs, so this is
            # not static. We will archive the whole working directory, but we can
            # use this file list to do additional checks for task success and allow
            # tighter output binding.
            "output": list(outputs),
        },
    }
    task_node["result"] = task_node["uid"]

    output_node: dict = {
        "object_type": "ExecutableOutput",
        "uid": _next_uid().hex(),
        "input": {
            # Explicit dependency that is not (yet) resolved in terms of data flow.
            "depends": task_node["uid"],
            # Fields with dependencies, though not evident in this example.
            "outputs": list(outputs),
            "stdout": stdout,
            "stderr": stderr,
        },
    }
    output_node["result"] = output_node["uid"]

    return {
        "implementation": ["scalems", "wrappers", "gromacs", "Executable"],
        "message": {"Executable": [input_node, task_node, output_node]},
    }


try:
    from gmxapi.commandline import cli_executable

    _gmx_cli_entry_point = cli_executable()
except ImportError:
    cli_executable = None
    _gmx_cli_entry_point = None
if _gmx_cli_entry_point is None:
    _gmx_cli_entry_point = "gmx"


# TODO: Implicit ensemble handling requires type annotations.
# def make_input(simulation_parameters=['md.mdp'],
#                topology=['md.top'],
#                conformation=['md.gro'],
#                wrapper_name='gmx'):
def make_input(
    simulation_parameters="md.mdp", topology="md.top", conformation="md.gro", wrapper_name=_gmx_cli_entry_point
):
    preprocess = _executable(
        argv=(
            wrapper_name,
            "grompp",
            "-f",
            simulation_parameters,
            "-p",
            topology,
            "-c",
            conformation,
            "-o",
            "topol.tpr",
        )
    )
    return {
        "implementation": ["scalems", "wrappers", "gromacs", "SimulationInput"],
        "message": {"SimulationInput": {"input": preprocess}},
    }


def simulate(arg, **kwargs):
    if not isinstance(arg, dict) or "SimulationInput" not in arg.get("message", ()):
        raise TypeError("Bad input.")
    simulator_input = arg["message"]["SimulationInput"]["input"]
    # Note: We haven't implemented modify_input or simulation chaining yet.
    if not isinstance(simulator_input, dict) or "implementation" not in simulator_input:
        raise TypeError("Wrong kind of input object.")
    return {
        "implementation": ["scalems", "wrappers", "gromacs", "Simulate"],
        "message": {"Simulate": {"input": simulator_input, "kwargs": kwargs}},
    }


def modify_input(*args, **kwargs):
    raise MissingImplementationError()


@functools.singledispatch
def scalems_helper(*args, **kwargs):
    """Allow SCALE-MS to manage tasks from this module.

    Placeholder module entry point. This explicit entry point will probably not
    be required (or appropriate) in the long run. Its roles can be subsumed by
    code that is already run during module import, such as by registration
    through decorators and/or subclassing.

    TODO:
        We can incorporate some amount of Capabilities Provider Interface (CPI)
        here to allow the module author to provide information to the workflow manager
        or dispatch along different run time code paths according to the execution environment.
    """
    logger.debug(
        "scalems_helper called with args:({}), kwargs:({})".format(
            ",".join([repr(arg) for arg in args]),
            ",".join([":".join((key, repr(value))) for key, value in kwargs.items()]),
        )
    )


@scalems_helper.register
def _(task: scalems.workflow.Task, context: WorkflowManager):
    sublogger = logger.getChild("scalems_helper")
    sublogger.debug("Serialized task record: {}".format(task.serialize()))
    command = task.type[-1]
    assert command in {"Simulate", "SimulationInput", "Executable"}
    # Right now, we should only need to process the 'Simulate' command...
    assert command == "Simulate"
    # TODO: Typing on the Task data proxy.
    command_message = task.input["message"][command]
    # kwargs = command_message['kwargs']
    input_ref: str = command_message["input"]
    logger.debug(f"Decoding reference {input_ref}")
    input_ref: bytes = bytes.fromhex(input_ref)
    task_map = context.task_map
    logger.debug(
        "Items done: {}".format(
            ", ".join([": ".join([key.hex(), str(value.done())]) for key, value in task_map.items()])
        )
    )
    # simulator_input_view = context.item(input_ref)

    # TODO: Workaround until we have the framework deliver results.
    # simulator_input: SubprocessResult = simulator_input_view.result()
    # logger.debug(f'Acquired grompp output: {simulator_input}')
    # tprfile = simulator_input.file['-o']
    dependency_dir = pathlib.Path(os.getcwd()).parent / input_ref.hex()
    tprfile = dependency_dir / "topol.tpr"
    import gmxapi

    md_in = gmxapi.read_tpr(str(tprfile))
    # TODO: Manage or integrate with gmxapi working directories.
    md = gmxapi.mdrun(md_in)
    # TODO: Make sure we mark interrupted simulations as "failed".
    return md.run()



"""Test dependency-management wrappers for RP.

Note: `export RADICAL_LOG_LVL=DEBUG` to enable RP debugging output.
"""
import asyncio
import logging

import pytest

import scalems
import scalems.context
import scalems.radical
import scalems.radical.runtime

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


# TODO: Catch sigint from RP and apply our own timeout.


@pytest.mark.asyncio
async def test_rp_future(rp_task_manager):
    """Check our Future implementation.

    Fulfill the asyncio.Future protocol for a rp.Task wrapper object. The wrapper
    should appropriately yield when the rp.Task is not finished.
    """
    import radical.pilot as rp

    timeout = 120

    tmgr = rp_task_manager

    task_description_dict = dict(
        executable="/bin/bash",
        arguments=["-c", "/bin/sleep 5 && /bin/echo success"],
        cpu_processes=1,
    )
    task_description = rp.TaskDescription(from_dict=task_description_dict)
    task_description.uid = "test-rp-future-1"

    # Test propagation of RP cancellation behavior
    task: rp.Task = tmgr.submit_tasks(task_description)

    rp_future: asyncio.Future = await scalems.radical.runtime.rp_task(task)

    task.cancel()
    try:
        with pytest.raises(asyncio.CancelledError):
            await asyncio.wait_for(rp_future, timeout=timeout)
    except asyncio.TimeoutError as e:
        # Useful point to insert an easy debugging break point
        raise e

    assert rp_future.cancelled()
    assert task.state == rp.states.CANCELED

    # Test propagation of asyncio watcher task cancellation.
    task_description.uid = "test-rp-future-2"
    task: rp.Task = tmgr.submit_tasks(task_description)

    rp_future: asyncio.Task = await scalems.radical.runtime.rp_task(task)

    assert isinstance(rp_future, asyncio.Task)
    rp_future.cancel()
    try:
        with pytest.raises(asyncio.CancelledError):
            await asyncio.wait_for(rp_future, timeout=5)
    except asyncio.TimeoutError as e:
        # Useful point to insert an easy debugging break point
        raise e
    assert rp_future.cancelled()

    # WARNING: rp.Task.wait blocks, and may never complete. Don't do it in the event loop thread.
    watcher = asyncio.create_task(asyncio.to_thread(task.wait, timeout=timeout), name=f"watch_{task.uid}")
    try:
        state = await asyncio.wait_for(watcher, timeout=timeout)
    except asyncio.TimeoutError as e:
        logger.exception(f"Waited more than {timeout} for {watcher}")
        watcher.cancel("Canceled after waiting too long.")
        raise e
    else:
        assert state in rp.states.FINAL
    # Note that if the test is paused by a debugger, the rp task may
    # have a chance to complete before being canceled.
    # Ordinarily, that will not happen in this test.
    # assert task.state in (rp.states.CANCELED, rp.states.DONE)
    assert task.state in (rp.states.CANCELED,)

    # Test run to completion
    task_description.uid = "test-rp-future-3"
    watcher = asyncio.create_task(asyncio.to_thread(tmgr.submit_tasks, task_description), name="rp_submit")
    try:
        task: rp.Task = await asyncio.wait_for(watcher, timeout=timeout)
    except asyncio.TimeoutError as e:
        logger.exception(f"Waited more than {timeout} to submit {task_description}.")
        watcher.cancel()
        raise e

    rp_future: asyncio.Task[rp.Task] = await scalems.radical.runtime.rp_task(task)
    try:
        result: rp.Task = await asyncio.wait_for(rp_future, timeout=timeout)
    except asyncio.TimeoutError as e:
        logger.debug(f"Waited more than {timeout} for {rp_future}: {e}")
        rp_future.cancel("Canceled after waiting too long.")
        raise e
    assert task.exit_code == 0
    assert "success" in task.stdout

    assert hasattr(result, "stdout")
    assert "success" in result.stdout
    assert rp_future.done()

    # Test failure handling
    # TODO: Use a separate test for results and error handling.


@pytest.mark.skip(reason="Unimplemented.")
@pytest.mark.asyncio
async def test_chained_commands():
    """Run a sequence of two tasks with a data flow dependency."""
    assert False



"""Manage non-volatile data for SCALE-MS in a filesystem context.

Implement several interfaces from `scalems.file`.

The data store must be active before attempting any manipulation of the managed workflow.
The data store must be closed by the time control returns to the interpreter after
leaving a managed workflow scope (and releasing a WorkflowManager instance).

We prefer to do this with the Python context manager protocol, since we already rely on
`scalems.workflow.scope()`.

We can also use a contextvars.ContextVar to hold a weakref to the FileStore, and register
a finalizer to perform a check, but the check could come late in the interpreter shutdown
and we should not rely on it. Also, note the sequence with which module variables and
class definitions are released during shutdown.
"""

from __future__ import annotations

__all__ = (
    "StaleFileStore",
    "filestore_generator",
    "get_file_reference",
    "FileReference",
    "FileStore",
    "FileStoreManager",
    "FilesView",
    "Metadata",
)

import asyncio
import contextvars
import dataclasses
import functools
import json
import logging
import os
import pathlib
import shutil
import tempfile
import threading
import typing
import weakref

import scalems.exceptions as _exceptions
from . import _lock
from .. import file as _file
from .. import identifiers as _identifiers

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))

_data_subdirectory = "scalems_0_0"
"""Subdirectory name to use for managed filestore."""

_metadata_filename = "scalems_context_metadata.json"
"""Name to use for Context metadata files (module constant)."""


class StaleFileStore(_exceptions.ContextError):
    """The backing file store is not consistent with the known (local) history."""


@dataclasses.dataclass
class Metadata:
    """Simple container for metadata at run time.

    This is intentionally decoupled from the non-volatile backing store or any usage
    protocols. Instances are managed through the FileStore.

    We will probably prefer to provide distinct interfaces for managing non-volatile
    assets (files) and for persistent workflow state (metadata), which may ultimately be
    supported by a single FileStore. That would likely involve replacing rather than
    extending this class, so let's keep it as a simple container for now:
    Make sure that this dataclass remains easily serialized and deserialized.

    For simplicity, this data structure should be easily serializable. Where necessary,
    mappings to and from non-trivial types must be documented.
    """

    instance: int
    """Unique identifier for this data store."""

    files: typing.MutableMapping[str, str] = dataclasses.field(default_factory=dict)
    """Managed filesystem objects.

    Maps hexadecimal-encoded (str) ResourceIdentifiers to filesystem paths (str).
    """

    objects: typing.MutableMapping[str, dict] = dataclasses.field(default_factory=dict)
    """Dictionary-encoded objects.

    Refer to `scalems.serialization` for encoding schema.
    """


class FilesView(typing.Mapping[_identifiers.ResourceIdentifier, pathlib.Path]):
    """Read-only viewer for files metadata.

    Provides a mapping from :py:class:`~scalems.identifiers.ResourceIdentifier` to
    the stored filesystem path.

    Warnings:
        The :py:class:`~pathlib.Path` objects represent paths in the context of
        the filesystem of the FileStore. If the FileStore does not represent a local
        filesystem, then a ``file://`` URIs from ``value.as_uri()`` is specific
        to the remote filesystem.
    """

    def __init__(self, files: typing.Mapping[str, str]):
        # Get a strong reference to the data structure provided by the filestore.
        self._files = files

    def __getitem__(self, k: _identifiers.ResourceIdentifier | str) -> pathlib.Path:
        return pathlib.Path(self._files[str(k)])

    def __contains__(self, k: _identifiers.ResourceIdentifier | str) -> bool:
        return str(k) in self._files

    def __len__(self) -> int:
        return len(self._files)

    def __iter__(self) -> typing.Iterator[_identifiers.ResourceIdentifier]:
        for key in self._files:
            yield _identifiers.ResourceIdentifier(bytes.fromhex(key))


class FileStore(typing.Mapping[_identifiers.ResourceIdentifier, "FileReference"]):
    """Handle to the SCALE-MS nonvolatile data store for a workflow context.

    Used as a Container, serves as a Mapping from identifiers to resource references.

    TODO: Consider generalizing the value type of the mapping for general Resource types, per the scalems data model.
    """

    _fields: typing.ClassVar = set([field.name for field in dataclasses.fields(Metadata)])
    _instances: typing.ClassVar = weakref.WeakValueDictionary()

    _token: contextvars.Token = None
    _data: Metadata
    _datastore: pathlib.Path = None
    _directory: pathlib.Path = None
    _filepath: pathlib.Path = None
    _update_lock: threading.Lock
    _dirty: threading.Event
    _log: typing.Sequence[str]
    _repr: str = "<FileStore>"

    @property
    def _tmpfile_prefix(self) -> str:
        """Filename generation helper for internal use."""
        return f"tmp_{self.instance}_"

    @property
    def log(self):
        """Get access to the operation log.

        Interface TBD...
        """
        for line in self._log:
            yield line

    # TODO: Consider a caching proxy to the directory structure to reduce filesystem
    #  calls.

    @property
    def directory(self) -> pathlib.Path:
        """The work directory under management.

        Generally, this is the current working directory when a scalems script is
        launched. The user is free to use the work directory to stage input and output
        files with the exception of the single scalems datastore
        directory. See `FileStore.datastore`.

        scalems will create a hidden "lock" directory briefly when starting up or
        shutting down. You may need to manually remove the lock directory after a
        particularly bad crash.
        """
        return self._directory

    @property
    def datastore(self) -> pathlib.Path:
        """Path to the data store for the workflow managed at *directory*.

        scalems creates a subdirectory at the root of a workflow in which to manage
        internal data. Editing files or directories in this subdirectory will affect
        workflow state and validity.

        The name reflects the SCALE-MS data format version and is not user-configurable.
        """
        if self._datastore is None:
            self._datastore = self.directory / _data_subdirectory
        return self._datastore

    @property
    def filepath(self) -> pathlib.Path:
        """Path to the metadata backing store.

        This is the metadata file used by scalems to track workflow state and
        file-backed data. Its format is closely tied to the scalems API level.
        """
        if self._filepath is None:
            self._filepath = self.datastore / _metadata_filename
        return self._filepath

    @property
    def instance(self):
        return self._data.instance

    def __repr__(self):
        return str(self._repr)

    def __getitem__(self, __k: _identifiers.ResourceIdentifier | str) -> "FileReference":
        return FileReference(filestore=self, key=__k)

    def __len__(self) -> int:
        return len(self._data.files)

    def __iter__(self) -> typing.Iterator[_identifiers.ResourceIdentifier]:
        return iter(self.files)

    def __init__(self, *, directory: pathlib.Path):
        """Assemble the data structure.

        Users should not create FileStore objects directly, but with
        initialize_datastore() or through the WorkflowManager instance.

        Once initialized, caller is responsible for calling the close() method.
        The easiest way to do this is to avoid creating the FileStore directly,
        and instead use a FileStoreManager object.

        No directory in the filesystem should be managed by more than one FileStore.
        The FileStore class maintains a registry of instances to prevent instantiation of
        a new FileStore for a directory that is already managed.

        Raises:
            ContextError: if attempting to instantiate for a directory that is already
                managed.
        """
        self._directory = pathlib.Path(directory).resolve()

        # TODO: Use file backing for journaled operations.
        self._log = []

        self._update_lock = threading.Lock()
        self._dirty = threading.Event()

        try:
            with _lock.scoped_directory_lock(self._directory):
                existing_filestore = FileStore._instances.get(self._directory, None)
                if existing_filestore is not None:
                    raise _exceptions.ContextError(f"{directory} is already managed by {repr(existing_filestore)}")
                # The current interpreter process is not aware of an instance for
                # *directory*

                for parent in self._directory.parents:
                    if _lock.is_locked(parent) or parent.joinpath(_data_subdirectory).exists():
                        raise _exceptions.ContextError(
                            f"Cannot establish scalems work directory {directory} "
                            f"because it is nested in work directory {parent}."
                        )

                instance_id = os.getpid()

                try:
                    self.datastore.mkdir()
                except FileExistsError:
                    # Check if the existing filesystem state is due to a previous clean
                    # shutdown, a previous dirty shutdown, or inappropriate concurrent
                    # access.
                    if not self.filepath.exists():
                        raise _exceptions.ContextError(
                            f"{self._directory} contains invalid datastore {self.datastore}."
                        )
                    logger.debug("Restoring metadata from previous session.")
                    with self.filepath.open() as fp:
                        metadata_dict = json.load(fp)
                    if metadata_dict["instance"] is not None:
                        # The metadata file has an owner.
                        if metadata_dict["instance"] != instance_id:
                            # This would be a good place to check a heart beat or
                            # otherwise
                            # try to confirm whether the previous owner is still active.
                            raise _exceptions.ContextError("Context is already in use.")
                        else:
                            assert metadata_dict["instance"] == instance_id
                            raise _exceptions.InternalError(
                                "This process appears already to be managing "
                                f"{self.filepath}, but is not "
                                "registered in FileStore._instances."
                            )
                    else:
                        # The metadata file has no owner. Presume clean shutdown
                        logger.debug(f"Taking ownership of metadata file for PID {instance_id}")
                        metadata_dict["instance"] = instance_id
                else:
                    logger.debug(f"Created new data store {self.datastore}.")
                    metadata_dict = {"instance": instance_id}

                metadata = Metadata(**metadata_dict)
                with open(self.filepath, "w") as fp:
                    json.dump(dataclasses.asdict(metadata), fp)
                FileStore._instances[self._directory] = self
                self.__dict__["_data"] = metadata

        except _lock.LockException as e:
            raise _exceptions.ContextError(
                "Could not acquire ownership of working directory {}".format(directory)
            ) from e
        self._repr = "<{cls} pid={pid} dir={dir}>".format(
            cls=self.__class__.__qualname__, pid=self.instance, dir=self.directory
        )
        assert FileStore._instances.get(self.directory) is self
        logger.debug(f"FileStore {self} open.")

    def flush(self):
        """Write the current metadata to the backing store, if there are pending
        updates.

        For atomic updates and better forensics, we write to a new file and *move* the
        file to replace the previous metadata file. If an error occurs, the temporary
        file will linger.

        Changes to metadata must be
        made with the update lock held and must end by setting the "dirty" condition
        (notifying the watchers of "dirty"). flush() should acquire the update lock and
        clear the "dirty" condition when successful. We can start with "dirty" as an
        Event. If we need more elaborate logic, we can use a Condition.
        It should be sufficient to use asyncio primitives, but we may need to use
        primitives from the `threading` module to allow metadata updates directly
        from RP task callbacks.

        """
        with self._update_lock:
            if self._dirty.is_set():
                # Warning: This can throw if, for instance, the parent directory disappears.
                with tempfile.NamedTemporaryFile(
                    dir=self.datastore, delete=False, mode="w", prefix="flush_", suffix=".json"
                ) as fp:
                    json.dump(dataclasses.asdict(self._data), fp)
                if self.closed:
                    logger.error(f"Leaving temporary metadata file {fp.name} from attempt to flush stale {self}.")
                    raise StaleFileStore("Cannot flush stale FileStore.")
                pathlib.Path(fp.name).rename(self.filepath)
                self._dirty.clear()

    def close(self):
        """Flush, shut down, and disconnect the FileStore from the managed directory.

        Not robustly thread safe. User is responsible for serializing access, as necessary,
        though ideally *close()* should be called exactly once, after the instance is no
        longer in use.

        Raises:
            StaleFileStore if called on an invalid or outdated handle.
            ScopeError if called from a disallowed context, such as from a forked process.

        """
        actual_manager = FileStore._instances.get(self._directory, None)
        if actual_manager is None:
            # Since FileStore._instances is a WeakValueDict, the entry may already be
            # gone if this call to `close` is happening during garbage collection.
            logger.debug(f"FileStore._instances[{self.directory}] is already empty")
        else:
            if actual_manager is not self:
                raise StaleFileStore(f"{self.directory} is currently managed by {actual_manager}")

        try:
            self.flush()
        except Exception as e:
            raise StaleFileStore(f"Could not flush {repr(self)}.") from e

        current_instance = getattr(self, "instance", None)
        self._data.instance = None

        try:
            with _lock.scoped_directory_lock(self.directory):
                if FileStore._instances.get(self._directory, None) is self:
                    del FileStore._instances[self._directory]
                if current_instance is None:
                    raise StaleFileStore(f"{self} is already closed.")
                elif current_instance != os.getpid():
                    raise _exceptions.ScopeError("Calling close() on a FileStore from another process is not allowed.")
                with open(self.filepath, "r") as fp:
                    context = json.load(fp)
                stored_instance = context.get("instance", None)
                if stored_instance != current_instance:
                    # Note that the StaleFileStore check will be more intricate as
                    # metadata becomes more sophisticated.
                    raise StaleFileStore(
                        "Expected ownership by {}, but found {}".format(current_instance, stored_instance)
                    )
                else:
                    context["instance"] = None
                    with open(self.filepath, "w") as fp:
                        json.dump(context, fp)
        except OSError as e:
            raise _exceptions.ContextError("Trouble with workflow directory access during shutdown.") from e

        logger.debug(f"FileStore {self} closed.")

    @property
    def closed(self) -> bool:
        if self._data.instance is None:
            actual_manager = FileStore._instances.get(self.directory, None)
            if actual_manager is self:
                raise _exceptions.ContextError(f"{self} is in an invalid state.")
            else:
                return True
        return False

    async def add_file(
        self, obj: typing.Union[_file.AbstractFile, _file.AbstractFileReference], _name: str = None
    ) -> FileReference:
        """Add a file to the file store.

        Not thread safe. User is responsible for serializing access, as necessary.

        We require file paths to be wrapped in a special type so that we can enforce
        that some error checking is possible before the coroutine actually runs. See

        This involves placing (copying) the file, reading the file to fingerprint it,
        and then writing metadata for the file. For clarity, we also rename the file
        after fingerprinting to remove a layer of indirection, even though this
        generates additional load on the filesystem.

        The caller is allowed to provide a *_name* to use for the filename,
        but this is strongly discouraged. The file will still be fingerprinted and the
        caller must be sure the key is unique across all software components that may
        by using the FileStore. (This use case should be limited to internal use or
        core functionality to minimize namespace collisions.)

        If the file is provided as a memory buffer or io.IOBase subclass, further
        optimization is possible to reduce the filesystem interaction to a single
        buffered write, but such optimizations are not provided natively by
        FileStore.add_file(). Instead, such optimizations may be provided by utility
        functions that produce AbstractFileReference objects that FileStore.add_file() can
        consume.

        Raises:
            scalems.exceptions.DuplicateKeyError: if a file with the same
                `ResourceIdentifier` is already under management.
            StaleFileStore: if the filestore state appears invalid.

        In addition to TypeError and ValueError for invalid inputs, propagates
        exceptions raised by failed attempts to access the provided file object.

        See Also:
            :py:func:`scalems.file.describe_file()`

        """
        # Note: We can use a more effective and flexible `try: ... except: ...` check
        # once we actually have AbstractFileReference support ready to use here, at which point
        # we can consider removing runtime_checkable from AbstractFileReference to avoid false
        # confidence. However, we may prefer to convert AbstractFileReference to an abc.ABC so
        # that we can actually functools.singledispatchmethod on the provided object.
        if isinstance(obj, _file.AbstractFileReference):
            raise _exceptions.MissingImplementationError("AbstractFileReference input not yet supported.")
        # Early exit check. We need to check again once we have a lock.
        if self.closed:
            raise StaleFileStore("Cannot add file to a closed FileStore.")
        path: pathlib.Path = pathlib.Path(obj)
        filename = None
        key = None
        _fd, tmpfile_target = tempfile.mkstemp(prefix=self._tmpfile_prefix, dir=self.datastore)
        os.close(_fd)
        tmpfile_target = pathlib.Path(tmpfile_target)
        try:
            # 1. Copy the file.
            kwargs = dict(src=path, dst=tmpfile_target, follow_symlinks=False)
            _path = await asyncio.to_thread(shutil.copyfile, **kwargs)
            assert tmpfile_target == _path
            assert tmpfile_target.exists()
            # We have now secured a copy of the file. We could just schedule the remaining
            # operations and return. For the initial implementation though, we will wait
            # until we have read the file. When optimizing for inputs that are already
            # fingerprinted, we could return before AbstractFileReference.localize() has
            # finished and schedule a checksum verification that localize() must wait for.

            # 2. Fingerprint the file.
            checksum = await asyncio.to_thread(obj.fingerprint, **{})
            identifier = _identifiers.ResourceIdentifier(checksum)

            # 3. Choose a key.
            key = str(identifier)

            # 4. Write metadata.
            if _name is None:
                _name = key
            filename = self.datastore.joinpath(_name)
            with self._update_lock:
                # Check again, now that we have the lock.
                if self.closed:
                    raise StaleFileStore("Cannot add file to a closed FileStore.")

                if key in self._data.files:
                    raise _exceptions.DuplicateKeyError(f"FileStore is already managing a file with key {key}.")
                if str(filename) in self._data.files.values():
                    raise _exceptions.DuplicateKeyError(f"FileStore already has {filename}.")
                else:
                    if filename.exists():
                        raise StaleFileStore(f"Unexpected file in filestore: {filename}.")
                os.rename(tmpfile_target, filename)
                logger.debug(f"Placed {filename}")
                # TODO: Enforce the setting of the "dirty" flag.
                # We can create a stronger separation between the metadata manager and
                # the backing data store and/or restrict assignment to use a general
                # Data Descriptor that manages our "dirty" flag.
                self._dirty.set()
                self._data.files[key] = str(filename)

            return FileReference(filestore=self, key=key)
        except (_exceptions.DuplicateKeyError, StaleFileStore) as e:
            tmpfile_target.unlink(missing_ok=True)
            raise e
        except Exception as e:
            # Something went particularly wrong. Let's try to clean up as best we can.
            logger.exception(f"Unhandled exception while trying to store {obj}")
            if key and key in self._data.files:
                pathlib.Path(self._data.files[key]).unlink(missing_ok=True)
                del self._data.files[key]
            if isinstance(filename, pathlib.Path):
                filename.unlink(missing_ok=True)
            raise e
        finally:
            if tmpfile_target.exists():
                logger.warning(f"Temporary file left at {tmpfile_target}")

    @property
    def files(self) -> typing.Mapping[_identifiers.ResourceIdentifier, pathlib.Path]:
        """Proxy to the managed files metadata.

        The proxy is read-only, but dynamic. A reference to *FileStore.files*
        will provide a dynamic view for later use.

        Warning:
            The returned FilesView will extend the lifetime of the metadata structure,
            but will not extend the lifetime of the FileStore itself.
        """
        if self.closed:
            raise StaleFileStore("FileStore is closed.")
        return FilesView(self._data.files)

    def add_task(self, identity: _identifiers.Identifier, **kwargs):
        """Add a new task entry to the metadata store.

        Not thread-safe. (We may need to manage this in a thread-safe manner, but it is
        probably preferable to do that by delegating a single-thread or thread-safe manager
        to serialize metadata edits.)
        """
        # Design note: among other interface requirements, the Metadata interface
        # should adapt between scalems types and serializable forms, so that str
        # conversions like the following become inappropriate.
        if str(identity) in self._data.objects:
            # TODO: It is not an error to add an item with the same key if it
            #  represents the same workflow object (potentially in a more or less
            #  advanced state).
            raise _exceptions.DuplicateKeyError(f"Task {identity} is already in the metadata store.")
        else:
            # Note: we don't attempt to be thread-safe, and this is not (yet) an async
            # coroutine, so we don't bother with a lock at this point.
            self._dirty.set()
            self._data.objects[str(identity)] = dict(**kwargs)
            # TODO: Take care with the value types that we can serialize and
            #  deserialize them.


class FileReference(_file.AbstractFileReference):
    """Provide access to a :py:class:`File <scalems.file.AbstractFile>` managed by a `FileStore`.

    Implements :py:class:`~scalems.file.AbstractFileReference`.
    """

    def __init__(self, filestore: FileStore, key: _identifiers.ResourceIdentifier | str):
        self._filestore: FileStore = filestore
        if key not in self._filestore.files:
            raise _exceptions.ProtocolError("Cannot create reference to unmanaged file.")
        if not isinstance(key, _identifiers.ResourceIdentifier):
            key = _identifiers.ResourceIdentifier(bytes.fromhex(key))
        self._key = key
        self._repr = f"<FileReference key:{key} path:{self._filestore.files[key]}>"

    def __fspath__(self) -> str:
        if not self.is_local():
            raise _file.DataLocalizationError(
                f"Non-local object (managed by {self._filestore}) is not implicitly convertible to filesystem path."
            )
        return str(self._filestore.files[self._key])

    def is_local(self, context=None) -> bool:
        if context is None:
            context = self._filestore
        return self._key in context.files

    async def localize(self, context=None) -> "FileReference":
        if context is None:
            context = self._filestore
        if context is not self._filestore:
            raise _exceptions.MissingImplementationError("Push subscription not yet implemented.")
        assert self._key in context.files
        return self

    def filestore(self) -> FileStore:
        return self._filestore

    def key(self) -> _identifiers.ResourceIdentifier:
        return self._key

    def as_uri(self, context=None) -> str:
        if context is None:
            context = self._filestore
        if context is not self._filestore:
            raise _exceptions.MissingImplementationError("Path resolution dispatching is not yet implemented.")
        if self.is_local():
            return context.files[self._key].as_uri()
        else:
            raise _exceptions.MissingImplementationError("No URI available for non-local files.")

    def __repr__(self):
        return self._repr


def filestore_generator(directory=None):
    """Generator function to manage a FileStore instance.

    Initializes a FileStore and yields (a weak reference to) it an arbitrary number of
    times. When the generator is garbage collected, or otherwise interrupted, properly
    closes the FileStore before finishing.

    This is a simple "coroutine" using the extended generator functionality from PEP 342.
    See Also:
        https://docs.python.org/3/howto/functional.html#passing-values-into-a-generator

    """
    if directory is None:
        directory = os.getcwd()
    path = pathlib.Path(directory)
    filestore = FileStore(directory=path)
    _closed = filestore.closed
    try:
        while not filestore.closed:
            message = yield weakref.ref(filestore)
            # We do not yet have any protocols requiring us to send messages to the
            # filestore when getting it.
            assert message is None
            # Note that we could also do interesting things with nested generators/coroutines
            # Ref PEP-0380: "return expr in a generator causes StopIteration(expr)
            #                to be raised upon exit from the generator."
        _closed = True
    finally:
        # The expected use case is that the generator will be abandoned when the owner
        # of the FileStoreManager is destroyed, resulting in
        # the Python interpreter calling the `close()` method, raising a GeneratorExit
        # exception at the above `yield`. We don't need to catch that explicitly. We
        # just need to clean up when the filestore is no longer in use.
        try:
            if not _closed:
                logger.debug(f"filestore_generator is closing {filestore}.")
                filestore.close()
        except Exception:
            # Note that this exception will probably be in the context of handling the
            # GeneratorExit exception from the `try` block.
            logger.exception(f"Exception while trying to close {filestore}.")
        else:
            assert filestore.closed


class FileStoreManager:
    def __init__(self, directory=None):
        self.filestore_generator = filestore_generator(directory=directory)
        # Make sure generator function runs up to the first yield.
        next(self.filestore_generator)

    def filestore(self) -> typing.Union[None, FileStore]:
        try:
            ref = next(self.filestore_generator)
            return ref()
        except StopIteration:
            logger.error("Managed FileStore is no longer available.")
            return None

    def close(self):
        if self.filestore_generator is None:
            return False
        else:
            store = self.filestore()
            if store is not None:
                store.close()
            try:
                next(self.filestore_generator)
            except StopIteration:
                logger.debug(f"{self} has been closed.")
            self.filestore_generator = None
            return True


async def _get_file_reference_by_path(obj: pathlib.Path, *, filestore: FileStore, mode="rb") -> FileReference:
    """Get file reference, adding to filestore, if and only if necessary.

    TODO: Optimize. (This call should only read and fingerprint *obj* once.)
    """
    obj = _file.describe_file(obj.resolve(), mode=mode)
    try:
        file_ref = await filestore.add_file(_file.describe_file(obj, mode=mode))
    except _exceptions.DuplicateKeyError:
        # Return the existing fileref, if it matches the fingerprint of *obj*.
        fingerprint = await asyncio.to_thread(obj.fingerprint)
        key = _identifiers.ResourceIdentifier(fingerprint)
        file_ref = filestore.get(key)
    return file_ref


@functools.singledispatch
def get_file_reference(obj, filestore=None, mode="rb") -> typing.Coroutine[None, None, FileReference]:
    """Get a FileReference for the provided object.

    If *filestore* is provided, use the given FileStore to manage the `FileReference`.
    Otherwise, use the FileStore for the current WorkflowManager.

    This is a dispatching function. Handlers for particular object types must are
    registered by decorating with ``@get_file_reference.register``. See
    :py:func:`functools.singledispatch`.

    Args:
        obj: identify the (managed) file
        filestore: file resource management backing store
        mode (str) : file access mode

    If *obj* is a :py:class:`~pathlib.Path` or :py:class:`~os.PathLike` object,
    add the identified file to the file store. If *obj* is a `ResourceIdentifier`,
    attempts to retrieve a reference to the identified filesystem object.

    Note:
        If *obj* is a `str` or :py:class:`bytes`, *obj* is interpreted as a `ResourceIdentifier`,
        **not** as a filesystem path.

    TODO: Either try to disambiguate str and bytes, or disallow and require stronger typed inputs.

    This involves placing (copying) the file, reading the file to fingerprint it,
    and then writing metadata for the file. For clarity, we also rename the file
    after fingerprinting to remove a layer of indirection, even though this
    generates additional load on the filesystem.

    In addition to TypeError and ValueError for invalid inputs, propagates
    exceptions raised by failed attempts to access the provided file object.

    TODO: Try to detect file type. See, for instance, https://pypi.org/project/python-magic/
    """
    try:
        return filestore.get_file_reference(obj, mode=mode)
    except AttributeError:
        # We don't mind if *filestore* does not provide this method.
        pass
    # We might expect *filestore* to raise NotImplemented or TypeError if it is
    # unable to handle the dispatch. This would not be an error in itself, except that
    # we do not have any other fall-back dispatching for types that have not been
    # registered.
    raise TypeError(f"Cannot convert {obj.__class__.__qualname__} to FileReference.")


@get_file_reference.register
def _(obj: pathlib.Path, filestore=None, mode="rb") -> typing.Coroutine[None, None, FileReference]:
    if filestore is None:
        raise _exceptions.MissingImplementationError(
            'There is not yet a "default" data store. *filestore* must be provided.'
        )
    else:
        # TODO: Check whether filestore is local or whether we need to proxy the object.
        ...

    # Should we assume that a Path object is intended to refer to a local file? We don't
    # want to  be ambiguous if the same path exists locally and remotely.
    if not obj.exists() or not obj.is_file():
        raise ValueError(f"Path {obj} is not a valid local file.")

    return _get_file_reference_by_path(obj, filestore=filestore, mode=mode)


@get_file_reference.register
def _(obj: str, *args, **kwargs) -> typing.Coroutine[None, None, FileReference]:
    return get_file_reference(_identifiers.ResourceIdentifier(bytes.fromhex(obj)), *args, **kwargs)


@get_file_reference.register
def _(obj: bytes, *args, **kwargs) -> typing.Coroutine[None, None, FileReference]:
    return get_file_reference(_identifiers.ResourceIdentifier(obj), *args, **kwargs)


@get_file_reference.register
def _(obj: os.PathLike, *args, **kwargs) -> typing.Coroutine[None, None, FileReference]:
    return get_file_reference(pathlib.Path(obj), *args, **kwargs)



"""WorkflowManager and Item interfaces.

Information and resources supporting a defined workflow are managed in the scope
of a :py:class:`~scalems.workflow.WorkflowManager`.
All ScaleMS API calls and references occur in or between the scopes of `WorkflowManager` instances.
When the ScaleMS API is active, there is a notion of a "current" WorkflowContext.

When the scalems module is simply imported, the default context is able to
discover local workflow state from the working directory and manage a workflow
representation, but it is not able to dispatch the work directly for execution.

SCALE-MS optimizes data flow and data locality in part by attributing all
workflow references to well-defined scopes. Stateful API facilities, workflow
state, and scoped references are managed as WorkflowManager instances.

"Workflow Manager" is an important parameter in several cases.
Execution management tools like scalems.run(), scalems.wait(), and scalems.dispatch()
update the workflow managed in a particular scope, possibly by interacting with
other scopes. Commands for declaring work or data add items to specific instances
of workflow managers. Internally, SCALEMS explicitly refers to theses scopes as
*manager*, or *context*, depending on the nature of the code collaboration.
*manager* may appear as an optional parameter for user-facing functions to allow
a particular managed workflow to be specified.

This module supports `scope()` and `get_scope()` with internal module state.
These tools interact with the context management of the asynchronous dispatching,
but note that they are not thread-safe. scope() should not be used in
a coroutine except in the root coroutine of a Task or otherwise within the scope
of a contextvars.copy_context().run(). scalems will try to flag misuse by raising
a ProtocolError, but please be sensible.
"""

__all__ = ("WorkflowManager", "get_scope", "scope", "wait", "workflow_item_director_factory")

import asyncio
import contextlib
import contextvars
import dataclasses
import functools
import json
import logging
import os
import queue as _queue
import threading
import typing
import weakref

from scalems import exceptions as _exceptions

from scalems.store import FileStoreManager
from scalems.messages import CommandQueueAddItem
from scalems.messages import CommandQueueControlItem
from scalems.messages import QueueItem
from scalems.exceptions import APIError
from scalems.exceptions import DispatchError
from scalems.exceptions import DuplicateKeyError
from scalems.exceptions import InternalError
from scalems.exceptions import MissingImplementationError
from scalems.exceptions import ProtocolError
from scalems.exceptions import ScaleMSError
from scalems.exceptions import ScopeError
from scalems.identifiers import TypeIdentifier
from scalems.serialization import encode
from scalems.unique import next_monotonic_integer

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


class Description:
    def shape(self) -> tuple:
        return self._shape

    def type(self) -> TypeIdentifier:
        return self._type

    def __init__(self, resource_type: TypeIdentifier, *, shape: tuple = (1,)):
        type_id = TypeIdentifier.copy_from(resource_type)
        if type_id is None:
            raise ValueError(f"Could not create a TypeIdentifier for {repr(resource_type)}")
        # TODO: Further input validation
        self._shape = shape
        self._type = type_id


class ItemView:
    """Standard object returned by a WorkflowContext when adding details to the workflow.

    Provides the normative interface for workflow items as a user-space object
    that proxies access through a workflow manager.

    Provides a Future like interface.

    At least in the initial implementation, a ItemView does not extend the lifetime
    of the WorkflowManager to which it refers.
    If the WorkflowManager from which the ItemView was obtained goes out of scope or
    otherwise becomes invalid, some ItemView interfaces can raise exceptions.

    .. todo:: Allows proxied access to future results through attribute access.

    """

    _workflow_manager: weakref.ReferenceType["WorkflowManager"]

    def uid(self) -> bytes:
        """Get the canonical unique identifier for this task.

        The identifier is universally unique and can be used to query any
        workflow manager for awareness of the task and (if the context is aware
        of the task) to get a view of the task.

        Returns:
            256-bit binary digest as a 32 element byte sequence.
        """
        return bytes(self._uid)

    def manager(self) -> "WorkflowManager":
        manager = self._workflow_manager()
        if manager is None:
            raise ScopeError("Out of scope. Managing context no longer exists!")
        else:
            if not isinstance(manager, WorkflowManager):
                raise APIError("Bug: ItemView._workflow_manager must weakly reference a WorkflowManager.")
        return manager

    def done(self) -> bool:
        """Check the status of the task.

        Returns:
            true if the task has finished.

        """
        manager = self.manager()
        return manager.tasks[self.uid()].done()

    def result(self):
        """Get a local object of the tasks's result type.

        .. todo:: Forces dependency resolution.

        """
        manager = self.manager()
        # TODO: What is the public interface to the tasks or completion status?
        # Note: We want to keep the View object as lightweight as possible, such
        # as storing just the weak ref to the manager, and the item identifier.
        return manager.tasks[self.uid()].result()

    def description(self) -> Description:
        """Get a description of the resource type."""
        manager = self.manager()
        return manager.tasks[self.uid()].description()

    def __getattr__(self, item):
        """Proxy attribute accessor for special task members.

        If the workflow element provides the requested interface, the managing
        Context will provide appropriate access. For "result" attributes, the
        returned reference supports the Future interface.
        """
        # We don't actually want to do this check here, but this is essentially what
        # needs to happen:
        #     assert hasattr(self.description().type().result_description().type(), item)
        manager: WorkflowManager = self.manager()
        task = manager.tasks[self.uid()]  # type: Task
        try:
            return getattr(task, item)
        except KeyError as e:
            raise e

    def __init__(self, manager, uid: bytes):
        self._workflow_manager = weakref.ref(manager)
        if isinstance(uid, bytes) and len(uid) == 32:
            self._uid = uid
        else:
            raise ProtocolError(f"uid should be a 32-byte binary digest (bytes). Got {repr(uid)}")


class InvalidStateError(ScaleMSError):
    """Object state is not compatible with attempted access.

    May result from a temporarily unserviceable request or an inappropriate
    usage. Emitter should provide a specific message.

    This exception serves a similar function to asyncio.InvalidStateError.
    """


class Task:
    """Encapsulate the implementation details of a managed Task workflow item.

    * Provide the interface required to support ItemView.
    * Wrap the task execution handling details.

    Note: We need an interface to describe the return value of manager.item().
    We currently manage tasks internally via Task instances, but that is a detail
    that is subject to change without warning.

    Warning:
        Not thread-safe. Access should be mediated by WorkflowManager protocols
        within the asyncio event loop.

    Once Task objects are created, they may not be updated directly by the client.

    TODO:
        Really, this can and should be implemented in terms of asyncio.Task with
        some additional SCALEMS overlay.

    State machine:
        Initial (no coroutine) -> Dispatched (coroutine) -> Executing (Task or Future)
        -> Final (awaited)

    TODO:
        Allow composable details. A WorkflowManager can provide its own task factory,
        but it is convenient to provide a complete base implementation, and extension
        is likely limited to how the task is dispatched and how the Futures are
        fulfilled, which seem very pluggable.
    """

    def result(self):
        if not self.done():
            raise InvalidStateError("Called result() on a Task that is not done.")
        return self._result

    def description(self) -> Description:
        decoded_record = json.loads(self._serialized_record)
        resource_type = TypeIdentifier(tuple(decoded_record["type"]))
        # TODO: Handle multidimensional references.
        shape = (1,)
        value = Description(resource_type=resource_type, shape=shape)
        return value

    def uid(self) -> bytes:
        if not isinstance(self._uid, bytes):
            raise InternalError("Task._uid was stored as bytes. Implementation changed?")
        return bytes(self._uid)

    def done(self) -> bool:
        return self._done.is_set()

    def __getattr__(self, item):
        # TODO: Manage internal state updates.
        # TODO: Respect different attribute type semantics.
        try:
            decoded_record = json.loads(self._serialized_record)
            logger.debug("Decoded {}: {}".format(type(decoded_record).__name__, str(decoded_record)))
            value = decoded_record[item]
        except json.JSONDecodeError as e:
            raise AttributeError('Problem retrieving "{}"'.format(item)) from e
        except KeyError as e:
            logger.debug('Did not find "{}" in {}'.format(item, self._serialized_record))
            raise AttributeError('Problem retrieving "{}"'.format(item)) from e
        return value

    def __init__(self, manager: "WorkflowManager", record: str):
        self._serialized_record = str(record)
        decoded_record = json.loads(self._serialized_record)

        self._uid = bytes.fromhex(decoded_record["uid"])
        if not len(self._uid) == 256 // 8:
            raise ProtocolError("UID is supposed to be a 256-bit hash digest. " f"Got {repr(self._uid)}")
        self._done = asyncio.Event()
        self._result = None

        # As long as we are storing Tasks in the workflow, we cannot store
        # workflows in Tasks.
        self._context = weakref.ref(manager)

    def set_result(self, result):
        # Not thread-safe.
        if self._done.is_set():
            raise ProtocolError("Result is already set for {}.".format(repr(self)))
        self._result = result
        self._done.set()
        logger.debug("Result set for {} in {}".format(self.uid().hex(), str(self._context())))

    def manager(self):
        return self._context()

    # @classmethod
    # def deserialize(cls, context, record: str):
    #     item_view = context.add_item(record)
    #     return item_view

    def serialize(self):
        return self._serialized_record


class TaskMap(dict, typing.MutableMapping[bytes, Task]):
    def __setitem__(self, k: bytes, v: Task) -> None:
        if not isinstance(k, bytes) and isinstance(k, typing.SupportsBytes):
            k = bytes(k)
        if not isinstance(k, bytes) or not isinstance(v, Task):
            raise TypeError("TaskMap maps *bytes* keys to *Task* values.")
        super().__setitem__(k, v)


_QItem_T = typing.TypeVar("_QItem_T", bound=QueueItem)
AddItemCallback = typing.Callable[[_QItem_T], None]


class WorkflowManager:
    """Composable context for SCALE-MS workflow management.

    A workflow manager includes a strategy for dispatching a workflow
    for execution. Instances provide the concurrent.futures.Executor
    interface with support and semantics that depend on the Executor
    implementation and execution environment.

    Notably, we rely on the Python contextmanager protocol to regulate
    the acquisition and release of resources, so SCALE-MS workflow
    contexts do not initialize Executors at creation. Instead,
    client code should use `with` blocks for scoped initialization and
    *shutdown* of Executor roles.

    .. todo:: Enforce centralization of Context instantiation for the interpreter process.

    For instance,

    * Implement a root context singleton and require acquisition of new Context
      handles through methods in this module.
    * Use abstract base class machinery to register Context implementations.
    * Require Context instances to track their parent Context, or otherwise
      participate in a single tree structure.
    * Prevent instantiation of Command references without a reference to a Context
      instance.

    TODO:
        In addition to adding callbacks to futures, allow subscriptions to workflow
        updates.
        This allows intermediate updates to be propagated and could be a superset of
        the additem hook.

    Notes:

        workflow manager maintains the workflow state without expensive or stateful
        volatile resources, and can mediate updates to the managed workflow at any
        time. Items enter the graph in an IDLE state. The WorkflowManager can provide
        Futures for the results of the managed items. For IDLE items,
        the WorkflowManager retains a weakref to the issued Futures, which it can use
        to make sure that there is only zero or one Future for a particular result.

        WorkflowManager collaborates with Queuer to transition the graph to an "active"
        or "executing" state. This transition is mediated through the dispatcher_lock.

        Queuer sequences and queues workflow items to be handled, pushing them to a
        dispatch_queue. No state change to the workflow item seems necessary at this
        time.

        The dispatch_queue is read by an ExecutionManager. Items may be processed
        immediately or staged in a command_queue. Workflow items are then either
        SUBMITTED or BLOCKED (awaiting dependencies). Optionally, Items may be marked
        ELIGIBLE and re-queued for batch submission.

        If the ExecutionManager is able to submit a task, the Task has a call-back
        registered for the workflow item. The WorkflowManager needs to convert any
        Future weakrefs to strong references when items are SUBMITTED, and the workflow
        Futures are subscribed to the item. Tasks are wrapped in a scalems object that
        the WorkflowManager is able to take ownership of. BLOCKED items are wrapped in
        Tasks which are subscribed to their dependencies (WorkflowItems should already
        be subscribed to WorkflowItem Futures for any dependencies) and stored by the
        ExecutionManager. When the call-backs for all of the dependencies indicate the
        Item should be processed into an upcoming workload, the Item becomes ELIGIBLE,
        and its wrapper Task (in collaboration with the ExecutionManager) puts it in
        the command_queue.

        As an optimization, and to support co-scheduling, a WorkflowItem call-back can
        provide notification of state changes. For instance, a BLOCKED item may become
        ELIGIBLE once all of its dependencies are SUBMITTED, when the actual Executor
        has some degree of data flow management capabilities.

    """

    tasks: TaskMap
    _executor_factory: typing.Callable
    """Factory for a RuntimeManager or subclass.

    TODO: Resolve circular reference between `execution` and `workflow` modules.
    """

    # TODO: Consider a threading.Lock for editing permissions.
    # TODO: Consider asyncio.Lock instances for non-thread-safe state updates during
    #  execution and dispatching.

    def __init__(
        self, *, loop: asyncio.AbstractEventLoop, executor_factory, directory: typing.Union[str, os.PathLike] = None
    ):
        """
        The event loop for the program should be launched in the root thread,
        preferably early in the application launch.
        Whether the WorkflowManager uses it directly,
        it is useful to require the client to provide the event loop,
        if for no other reason than to ensure that one exists.

        Args:
            loop: event loop, such as from asyncio.new_event_loop()
            executor_factory: Implementation-specific callable to get a run time work
                manager.
            directory: Filesystem path for the workflow file store.
        """
        # We are moving towards a composed rather than a derived WorkflowManager Context.
        # Note that we can require the super().__init__() to be called in derived classes,
        # so it is not non-sensical for an abc.ABC to have an __init__ method.
        if not isinstance(loop, asyncio.AbstractEventLoop):
            raise TypeError("Workflow manager requires an event loop object compatible with asyncio.AbstractEventLoop.")
        if loop.is_closed():
            raise ProtocolError("Event loop does not appear to be ready to use.")
        logger.debug(f"{repr(self)} acquired event loop {repr(loop)} at loop time {loop.time()}.")
        self._asyncio_event_loop = loop

        if not callable(executor_factory):
            raise TypeError("*executor_factory* argument must be a callable.")
        # TODO: Resolve circular reference between `execution` and `workflow` modules.
        # self._executor_factory: ExecutorFactory = executor_factory
        self._executor_factory = executor_factory

        # Basic Context implementation details
        # TODO: Tasks should only writable within a WorkflowEditor context.
        self.tasks = TaskMap()  # Map UIDs to task Futures.

        if directory is None:
            directory = os.getcwd()
        try:
            self._filestoremanager = FileStoreManager(directory=directory)
        except Exception as e:
            logger.exception(f"Could not initialize FileStoreManager for {directory}.", exc_info=e)
            raise ValueError("Need a usable local workflow *directory*.") from e
        # TODO: Restore workflow state from filestore.

        self._dispatcher: typing.Optional[Queuer] = None
        self._dispatcher_lock = asyncio.Lock()

        self._event_hooks: typing.Mapping[str, typing.MutableSet[AddItemCallback]] = {"add_item": set()}

    @property
    def closed(self):
        return self._filestoremanager is None

    def close(self):
        """Finalize and release resources.

        While not usually necessary, an explicit call to close() can force the backing
        data store to be closed and allow exceptions to be caught at a clear point.

        Once close() is called, additional attempts to access the managed workflow may
        raise ScopeError.
        """
        if self._dispatcher is not None:
            raise APIError("Cannot close a WorkflowManager that is actively dispatching work.")
        if self._filestoremanager is not None:
            self._filestoremanager.close()
            self._filestoremanager = None

    def loop(self):
        return self._asyncio_event_loop

    def subscribe(self, event: str, callback: AddItemCallback):
        """Register a callback for method call events.

        Currently, only the add_item method provides a subscribable event.

        Registered callbacks will be called before the subscribed-to method returns.
        The callback will be given a dictionary with the event name as a key
        and a value that is the same as the argument(s) to the method.

        Unsubscribe with unsubscribe() to avoid unexpected behavior and to avoid
        prolonging the life of the object providing the callback.
        """
        hook: typing.MutableSet[AddItemCallback] = self._event_hooks[event]
        # TODO: Do we need to allow the callback to include a contextvars.Context with
        #  the registration?
        if callback in hook:
            raise ValueError("Provided callback is already registered.")
        hook.add(callback)

    def unsubscribe(self, event: str, callback: AddItemCallback):
        hook = self._event_hooks[event]
        hook.remove(callback)

    def item(self, identifier) -> ItemView:
        """Access an item in the managed workflow."""
        # Consider providing the consumer context when acquiring access.
        # Consider limiting the scope of access requested.
        if isinstance(identifier, typing.SupportsBytes):
            identifier = bytes(identifier)
        if not isinstance(identifier, bytes):
            raise MissingImplementationError("Item look-up is currently limited to UID byte sequences.")
        identifier = bytes(identifier)
        logger.debug(
            "Looking for {} in ({})".format(identifier.hex(), ", ".join(key.hex() for key in self.tasks.keys()))
        )
        if identifier not in self.tasks:
            raise KeyError(f"WorkflowManager does not have item {str(identifier)}")
        item_view = ItemView(manager=self, uid=identifier)

        return item_view

    @contextlib.contextmanager
    def edit_item(self, identifier) -> typing.Generator[Task, None, None]:
        """Scoped workflow item editor.

        Grant the caller full access to the managed task.

        """
        if self.closed:
            raise ScopeError("WorkflowManager is closed.")
        # TODO: Use a proxy object that is easier to inspect and roll back.
        item = self.tasks[identifier]

        # We can add a lot more sophistication here.
        # For instance,
        # * we should lock items for editing
        # * we can update state metadata at each phase of the contextmanager protocol
        # * we can execute hooks based on the state change detected when the client
        #   releases the editor context.

        yield item

    # TODO: Consider helper functionality and `label` support.
    # def find(self, label: str = None):
    #     """Try to locate a workflow object.
    #
    #     Find reference by label. Find owner of non-local resource, if known.
    #     """

    def default_dispatcher(self):
        """Get a default dispatcher instance, if available.

        Provide a hint to scalems.run() on how to execute work in this scope.

        WorkflowManager implementations define their own semantics. If implemented,
        the returned object should be an AsyncContextManager. If the dispatching
        facility is not reentrant, the WorkflowManager may raise ProtocolError.

        WorkflowManagers are not required to provide a default dispatcher.
        """
        return None

    @contextlib.asynccontextmanager
    async def dispatch(self, dispatcher: "Queuer" = None, params=None):
        """Enter the execution dispatching state.

        Attach to a dispatching executor, then provide a scope for concurrent activity.
        This is also the scope during which the RADICAL Pilot Session exists.

        Provide the executor with any currently-managed work in a queue.
        While the context manager is active, new work added to the queue will be picked up
        by the executor. When the context manager is exited, new work will resume
        queuing locally and the remote tasks will be resolved, then the dispatcher
        will be disconnected.

        Currently, we tie the lifetime of the dispatcher to this context manager.
        When leaving the `with` block, we trigger the executor to clean-up and wait for
        its task to complete.
        We may choose some other relationship in the future.

        Args:
            dispatcher: A queue processor that will subscribe to the add_item hook to
            feed the executor.
            params: a parameters object relevant to the execution back-end

        .. todo:: Clarify re-entrance policy, thread-safety, etcetera, and enforce.

        """
        if self.closed:
            raise ScopeError("WorkflowManager is closed.")

        # 1. Bind a new executor to its queue.
        # 2. Bind a dispatcher to the executor.
        # 3. Enter executor context.
        # 4. Enter dispatcher context.
        #         # 1. (While blocking event loop in UI thread) Install a hook
        #              for the queuer to catch new calls to add_item (the
        #              dispatcher_queue).
        #         # 2. Get snapshot of current workflow state with which to initialize
        #              the executor. (Unblock.)
        #         # 3. Spool workflow snapshot to executor.
        #         # 4. Start dispatcher queue runner.
        #         # 5. Yield.
        # 5. Exit dispatcher context.
        # 6. Exit executor context.
        # TODO: Add lock context for WorkflowManager event hooks
        #  rather than assume the UI and event loop are always in the same thread.

        executor = self._executor_factory(manager=self, params=params)

        # Avoid race conditions while checking for a running dispatcher.
        # TODO: Clarify dispatcher state machine and remove/replace assertions.
        # Warning: The dispatching protocol is immature.
        # Initially, we don't expect contention for the lock,
        # and if there is contention, it probably represents
        # an unintended race condition or systematic dead-lock.
        assert not self._dispatcher_lock.locked()
        async with self._dispatcher_lock:
            # Dispatching state may be reentrant, but it does not make sense to
            # re-enter through this call structure.
            if self._dispatcher is not None:
                raise ProtocolError(f"Already dispatching through {repr(self._dispatcher)}.")
            if dispatcher is None:
                dispatcher = Queuer(source=self, command_queue=executor.queue(), dispatcher_lock=self._dispatcher_lock)
                self._dispatcher = dispatcher
            else:
                self._dispatcher = weakref.proxy(dispatcher)

        try:
            # Manage scope of executor operation with a context manager.
            # RP does not yet use an event loop, but we can use async context manager
            # for future compatibility with asyncio management of network connections,
            # etc.
            #
            # Note: the executor owns a rp.Session during operation.
            async with executor as dispatching_session:
                # Note: *executor* (sms.execution.RuntimeManager) returns itself when
                # "entered", then we yield it below. Now that RuntimeManager is
                # fairly normalized, we could pass the dispatcher to a (new)
                # context manager member function
                # and let the RuntimeManager handle all of this *dispatcher* logic.
                # The WorkflowManager could pass itself as a simpler interface
                # * to the Queuer for the `subscribe` add_item hook and
                # * to the RuntimeManager to provide a WorkflowEditor.edit_item.
                # E.g.
                #     @asynccontextmanager
                #     async def RuntimeManager.manage(
                #       dispatcher: Queuer,
                #       subscriber: WorkflowEditor))
                # Consider also the similarity of RuntimeManager-WorkflowManager-Queuer
                # to a Model-View-Controller.
                async with dispatcher:
                    # We can surrender control here and leave the executor and
                    # dispatcher tasks active while evaluating a `with` block suite
                    # for the `dispatch` context manager.
                    yield dispatching_session
                    # When leaving the `with` suite, Queuer.__aexit__ sends a *stop*
                    # command to the queue.
                # The *stop* command will be picked up by sms.execution.manage_execution()
                # (as the RuntimeManager's *runner_task*), which will be awaited in
                # RuntimeManager.__exit__().

        except Exception as e:
            logger.exception("Unhandled exception while in dispatching context.")
            raise e

        finally:
            # Warning: The dispatching protocol is immature.
            # Initially, we don't expect contention for the lock,
            # and if there is contention, it probably represents
            # an unintended race condition or systematic dead-lock.
            # TODO: Clarify dispatcher state machine and remove/replace assertions.
            #       Be on the look-out for nested context managers and usage in
            #       `finally` blocks.
            assert not self._dispatcher_lock.locked()
            async with self._dispatcher_lock:
                self._dispatcher = None

            dispatcher_exception = dispatcher.exception()
            if dispatcher_exception:
                if isinstance(dispatcher_exception, asyncio.CancelledError):
                    logger.info("Dispatching queue processor cancelled.")
                else:
                    assert not isinstance(dispatcher_exception, asyncio.CancelledError)
                    logger.exception("Queuer encountered exception.", exc_info=dispatcher_exception)
            else:
                if not dispatcher.queue().empty():
                    logger.error(
                        "Queuer finished while items remain in dispatcher queue. "
                        "Approximate size: {}".format(dispatcher.queue().qsize())
                    )

            executor_exception = executor.exception()
            if executor_exception:
                if isinstance(executor_exception, asyncio.CancelledError):
                    logger.info("Executor cancelled.")
                else:
                    assert not isinstance(executor_exception, asyncio.CancelledError)
                    logger.exception("Executor task finished with exception", exc_info=executor_exception)
            else:
                if not executor.queue().empty():
                    # TODO: Handle non-empty queue.
                    # There are various reasons that the queue might not be empty and
                    # we should clean up properly instead of bailing out or compounding
                    # exceptions.
                    # TODO: Check for extraneous extra *stop* commands.
                    logger.error("Bug: Executor left tasks in the queue without raising an exception.")

            logger.debug("Exiting {} dispatch context.".format(type(self).__name__))

    # @abc.abstractmethod
    # def add_task(self, operation: str, bound_input):
    #     """Add a task to the workflow.
    #
    #     Arguments:
    #         operation: named operation to perform
    #         bound_input: reference to a workflow object compatible with operation input.
    #
    #     Returns:
    #         Reference to the new task.
    #
    #     TODO: Resolve operation implementation to dispatch task configuration.
    #     """
    #     ...

    def add_item(self, item) -> ItemView:
        """
        Add an item to the managed workflow.

        Args:
            item: Describe the item to be managed. (Usually a task to be submitted.)

        Returns:
            View of the managed item.

        """
        if self.closed:
            raise ScopeError("WorkflowManager is closed.")

        # # TODO: Resolve implementation details for *operation*.
        # if operation != 'scalems.executable':
        #     raise MissingImplementationError('No implementation for {} in {}'.format(
        #     operation, repr(self)))
        # # Copy a static copy of the input.
        # # TODO: Dispatch tasks addition, allowing negotiation of Context capabilities
        #  and subscription
        # #  to resources owned by other Contexts.
        # if not isinstance(bound_input, scalems.subprocess.SubprocessInput):
        #     raise ValueError('Only scalems.subprocess.SubprocessInput objects
        #     supported as input.')

        # TODO: Replace with type-based dispatching or some normative interface test.
        from scalems.subprocess import Subprocess

        if not isinstance(item, (Subprocess, dict)):
            raise MissingImplementationError("Operation not supported.")

        # TODO: Generalize interface check.
        if isinstance(item, Subprocess):
            uid: bytes = item.uid()
            if uid in self.tasks:
                # TODO: Consider decreasing error level to `warning`.
                raise DuplicateKeyError("Task already present in workflow.")
            logger.debug("Adding {} to {}".format(str(item), str(self)))
            record = {"uid": uid.hex(), "type": item.resource_type().scoped_identifier(), "input": {}}
            task_input = item.input_collection()
            for field in dataclasses.fields(task_input):
                name = field.name
                try:
                    # TODO: Need serialization typing.
                    record["input"][name] = getattr(task_input, name)
                except AttributeError as e:
                    raise InternalError("Unexpected missing field.") from e
        else:
            # WARNING: no coverage
            assert isinstance(item, dict)
            assert "uid" in item
            uid = item["uid"]
            implementation_identifier = item.get("implementation", None)
            if not isinstance(implementation_identifier, list):
                raise DispatchError("Bug: bad schema checking?")

            if uid in self.tasks:
                # TODO: Consider decreasing error level to `warning`.
                raise DuplicateKeyError("Task already present in workflow.")
            logger.debug("Adding {} to {}".format(str(item), str(self)))
            record = {"uid": uid.hex(), "type": tuple(implementation_identifier), "input": item}
        serialized_record = json.dumps(record, default=encode)

        # TODO: Make sure there are no artifacts of shallow copies that may result in
        #       a user modifying nested objects unexpectedly.
        item = Task(self, serialized_record)
        # TODO: Check for ability to dispatch.
        #  Note module dependencies, etc. and check in target execution environment
        #  (e.g. https://docs.python.org/3/library/importlib.html#checking-if-a-module
        #  -can-be-imported)

        # TODO: Provide a data descriptor and possibly a more formal Workflow class.
        # We do not yet check that the derived classes actually initialize self.tasks.
        self.tasks[uid] = item

        task_view = ItemView(manager=self, uid=uid)

        # TODO: Register task factory (dependent on executor).
        # TODO: Register input factory (dependent on dispatcher and task factory /
        #  executor).
        # TODO: Register results handler (dependent on dispatcher end points).

        # TODO: Consider an abstract event hook for `add_item` and other (decorated)
        #  methods.
        # Internal functionality can probably explicitly register and unregister,
        # accounting for the current details of thread safety.
        # External access will need to be in terms of a concurrency framework,
        # so we can use a scoped `async with event_subscription`
        # to create an asynchronous iterator that a coroutine can use to receive
        # add_item messages
        # (with some means to externally end the subscription,
        # either through the generator protocol directly or through logic in the
        # provider of the iterator)
        for callback in self._event_hooks["add_item"]:
            # TODO: Do we need to provide a contextvars.Context object to the callback?
            logger.debug(f"Running dispatching hook for add_item subscriber {repr(callback)}.")
            callback(CommandQueueAddItem({"add_item": uid}))

        return task_view

    def datastore(self):
        """Get a handle to the non-volatile backing data store."""
        if self.closed:
            raise ScopeError("WorkflowManager is closed.")
        return self._filestoremanager.filestore()


@functools.singledispatch
def workflow_item_director_factory(
    item, *, manager: WorkflowManager, label: str = None
) -> typing.Callable[..., ItemView]:
    """Get a workflow item director for a workflow manager and input type.

    When called, the director finalizes the new item and returns a view.
    """
    # TODO: Incorporate into WorkflowManager interface as a singledispatchmethod.
    # Design note: WorkflowManager classes could cache implementation details for item
    # types,
    # but (since we acknowledge that work may be defined before instantiating the
    # context to
    # which it will be dispatched), we need to allow the WorkflowContext implementation to
    # negotiate/fetch the implementation details at any time.
    # In general, relationships between specific
    # workflow item types and context types should be resolved in terms of context traits,
    # not specific class definitions.
    # In practice, we have some more tightly-coupled relationships,
    # at least in early versions, as we handle
    # (a) subprocess-type items,
    # (b) function-based items,
    # and (c) data staging details for (1) local execution and (2) remote (RADICAL Pilot)
    # execution.
    raise MissingImplementationError("No registered implementation for {} in {}".format(repr(item), repr(manager)))


@workflow_item_director_factory.register
def _(item_type: type, *, manager: WorkflowManager, label: str = None) -> typing.Callable[..., ItemView]:
    # TODO: Do we really want to handle dispatching on type _or_ instance args?
    # When dispatching on a class instead of an instance, just construct an
    # object of the indicated type and re-dispatch. Note that implementers of
    # item types must register an overload with this singledispatch function.
    # TODO: Consider how best to register operations and manage their state.
    def constructor_proxy_director(*args, **kwargs) -> ItemView:
        if not isinstance(item_type, type):
            raise ProtocolError(
                "This function is intended for a dispatching path on which *item_type* is a `type` object."
            )
        item = item_type(*args, **kwargs)
        director = workflow_item_director_factory(item, manager=manager, label=label)
        return director()

    return constructor_proxy_director


class Queuer:
    """Maintain the active dispatching state for a managed workflow.

    The Queuer, WorkflowManager, and Executor lifetimes do not need to be
    coupled, but their periods of activity should be synchronized in certain ways
    (aided using the Python context manager protocol).

    The Queuer must have access to an active Executor while the Queuer
    is active.

    When entering the Queuer context manager, a task is created to transfer items
    from the dispatch queue to the execution queue. The task is allowed to drain the
    queue before the context manager yields to the ``with`` block. The task will
    continue to process the queue asynchronously if new items appear.

    When exiting the Queuer context manager, the items currently in the
    dispatching queue are processed and the the dispatcher task is finalized.

    The dispatcher queue is a queue.SimpleQueue so that WorkflowManager.add_item()
    can easily use a concurrency-safe callback to add items whether or not the
    dispatcher is active.
    """

    command_queue: asyncio.Queue
    """Target queue for dispatched commands."""

    source: WorkflowManager
    """Owner of the workflow items being queued."""

    _dispatcher_lock: asyncio.Lock
    """Provided by caller to allow safe transitions of dispatching state."""

    _dispatcher_queue: _queue.SimpleQueue

    _queue_runner_task: asyncio.Task
    """Queue, owned by this object, of Workflow Items being processed for dispatch."""

    def __init__(self, source: WorkflowManager, command_queue: asyncio.Queue, dispatcher_lock=None):
        """Create a queue-based workflow dispatcher.

        Initialization and deinitialization occurs through
        the Python (async) context manager protocol.
        """

        self.source = source
        self._dispatcher_queue = _queue.SimpleQueue()
        self.command_queue = command_queue

        if not isinstance(dispatcher_lock, asyncio.Lock):
            raise TypeError("An asyncio.Lock is required to control dispatcher state.")
        self._dispatcher_lock = dispatcher_lock

        self._exception = None

    def queue(self):
        return self._dispatcher_queue

    def put(self, item: typing.Union[CommandQueueAddItem, CommandQueueControlItem]):
        assert len(item) == 1
        key = list(item.keys())[0]
        if key not in {"command", "add_item"}:
            raise APIError("Unrecognized queue item representation.")
        self._dispatcher_queue.put(item)

    async def __aenter__(self):
        try:
            # Get a lock while the state is changing.
            # Warning: The dispatching protocol is immature.
            # Initially, we don't expect contention for the lock,
            # and if there is contention, it probably represents
            # an unintended race condition or systematic dead-lock.
            # TODO: Clarify dispatcher state machine and remove/replace assertions.
            assert not self._dispatcher_lock.locked()
            async with self._dispatcher_lock:
                if _dispatcher.get(None):
                    raise APIError("There is already an active dispatcher in this Context.")
                _dispatcher.set(self)
                # Launch queue processor (proxy executor).
                runner_started = asyncio.Event()
                runner_task = asyncio.create_task(self._queue_runner(runner_started))
                await runner_started.wait()
                self._queue_runner_task = runner_task

                # Without yielding,
                # 1. Install a hook for the queuer to catch new calls to add_item.
                # 2. Get snapshot of current workflow state with which to initialize
                #    the executor.
                # Dont' forget to unsubscribe later!
                # self.source_context.subscribe('add_item', self._dispatcher_queue.put)
                self.source.subscribe("add_item", self.put)
                # TODO: Topologically sort DAG!
                initial_task_list = list(self.source.tasks.keys())
                try:
                    for _task_id in initial_task_list:
                        self.command_queue.put_nowait(QueueItem({"add_item": _task_id}))
                except asyncio.QueueFull as e:
                    raise DispatchError("Executor was unable to receive initial commands.") from e
                # It is now safe to yield.

                # TODO: Add lock context for WorkflowManager event hooks
                #  rather than assume the UI and event loop are always in the same thread.

            return self
        except Exception as e:
            self._exception = e
            raise e

    async def _single_iteration_queue(self, source: _queue.SimpleQueue, target: asyncio.Queue):
        """Transfer one queue item.

        If a *stop* command is encountered, self-cancel after transfering command.

        To avoid race conditions while stopping queue processing,
        place a *stop* command in *source* and asyncio.shield() a call
        to this coroutine in a *try: ... except: ...* block.

        Note that the caller will then receive CancelledError after *stop* command has
        been transferred.

        Raises:
            queue.Empty if *source* is empty
            asyncio.CancelledError when cancelled or *stop* is received.

        """
        command: QueueItem = source.get_nowait()
        logger.debug(f"Processing command {repr(command)}")

        await target.put(command)

        # TODO: Use formal RPC protocol.
        if "control" in command:
            # Note that we don't necessarily need to stop managing the dispatcher queue
            # at this point, but the Executor will be directed to shut down,
            # so we must not put anything else onto the command queue until we have a
            # new command queue or a new executor.
            if command["control"] == "stop":
                raise asyncio.CancelledError()
            else:
                raise ProtocolError("Unknown command: {}".format(command["control"]))
        else:
            if "add_item" not in command:
                # TODO: We might want a call-back or Event to force errors before the
                #  queue-runner task is awaited.
                raise MissingImplementationError(f"Executor has no implementation for {str(command)}")
        return command

    async def _queue_runner(self, processing_state: asyncio.Event):
        processing_state.set()
        while True:
            try:
                await asyncio.shield(
                    self._single_iteration_queue(source=self._dispatcher_queue, target=self.command_queue)
                )
            except _queue.Empty:
                # Wait a moment and try again.
                await asyncio.sleep(0.5)

    async def _drain_queue(self):
        """Wait until the dispatcher queue is empty, then return.

        Use in place of join() for event-loop treatment of *queue.SimpleQueue*.
        """
        while not self._dispatcher_queue.empty():
            await asyncio.sleep(0.1)

    async def __aexit__(self, exc_type, exc_val, exc_tb):  # noqa: C901
        """Clean up at context exit.

        Drain the dispatching queue and exit.

        Unsubscribes from the WorkflowManager add_item hook, deactivates the
        dispatching context, and exits.
        Does not cancel or send instructions to the Executor managing the command queue.
        """
        # Note that this coroutine could take a long time and could be cancelled at
        # several points.
        cancelled_error = None
        # The dispatching protocol is immature. Initially, we don't expect contention
        # for the lock, and if there is contention, it probably represents an
        # unintended race condition or systematic dead-lock.
        # TODO: Clarify dispatcher state machine and remove/replace assertions.
        assert not self._dispatcher_lock.locked()
        async with self._dispatcher_lock:
            try:
                self.source.unsubscribe("add_item", self.put)
                _dispatcher.set(None)

                # Stop the dispatcher.
                logger.debug("Stopping the SCALEMS RP dispatching queue runner.")

                # Wait for the queue to drain or the queue runner to exit or fail.
                drain = asyncio.create_task(self._drain_queue())
                done, pending = await asyncio.wait(
                    {drain, self._queue_runner_task}, return_when=asyncio.FIRST_COMPLETED
                )
                assert len(done) > 0
                if self._queue_runner_task not in done:
                    if drain in done:
                        self._queue_runner_task.cancel()
                done, _ = await asyncio.wait({self._queue_runner_task})
                assert self._queue_runner_task in done
                if not self._queue_runner_task.cancelled():
                    exception = self._queue_runner_task.exception()
                else:
                    exception = None
                if exception:
                    logger.exception("Queuer queue processing encountered exception", exc_info=exception)
                    if self._exception:
                        logger.error("Queuer is already holding an exception.")
                    else:
                        self._exception = exception

            except asyncio.CancelledError as e:
                logger.debug("Queuer context manager received cancellation while exiting.")
                cancelled_error = e
            except Exception as e:
                logger.exception("Exception while stopping dispatcher.", exc_info=e)
                if self._exception:
                    logger.error("Queuer is already holding an exception.")
                else:
                    self._exception = e
            finally:
                # Should we do any other clean-up here?
                ...
        if cancelled_error:
            raise cancelled_error

        # Only return true if an exception should be suppressed (because it was handled).
        # TODO: Catch internal exceptions for useful logging and user-friendliness.
        if exc_type is not None:
            return False

    def exception(self) -> typing.Union[None, Exception]:
        return self._exception

    # We don't currently have a use for a stand-alone Task.
    # We use the async context manager and the exception() method.
    # def __await__(self):
    #     """Implement the asyncio task represented by this object."""
    #     # Note that this is not a native coroutine object; we cannot `await`
    #     # The role of object.__await__() is to return an iterator that will be
    #     # run to exhaustion in the context of an event loop.
    #     # We assume that most of the asyncio activity happens through the
    #     # async context mananager behavior and other async member functions.
    #     # If we choose to `await instance` at all, we need a light-weight
    #     # iteration we can perform to surrender control of the event loop,
    #     # and then just do some sort of tidying or reporting that doesn't fit well
    #     # into __aexit__(), such as the ability to return a value.
    #
    #     # # Note: We can't do this without first wait on some sort of Done event...
    #     # failures = []
    #     # for t in self.rp_tasks:
    #     #     logger.info('%s  %-10s : %s' % (t.uid, t.state, t.stdout))
    #     #     if t.state != rp.states.DONE or t.exit_code != 0:
    #     #         logger.error(f'RP Task unsuccessful: {repr(t)}')
    #     #         failures.append(t)
    #     # if len(failures) > 0:
    #     #     warnings.warn('Unsuccessful tasks: ' + ', '.join([repr(t) for t in
    #     failures]))
    #
    #     yield
    #     if self._exception:
    #         raise self._exception
    #
    #     # # If we want to provide a "Future-like" interface, we should support the
    #     callback
    #     # # protocols and implement the following generator function.
    #     # if not self.done():
    #     #     self._asyncio_future_blocking = True
    #     #     # ref https://docs.python.org/3/library/asyncio-future.html#asyncio
    #     .isfuture
    #     #
    #     #     yield self  # This tells Task to wait for completion.
    #     # if not self.done():
    #     #     raise RuntimeError("The dispatcher task was not 'await'ed.")
    #     # Ref PEP-0380: "return expr in a generator causes StopIteration(expr)
    #     # to be raised upon exit from the generator."
    #     # The Task works like a `result = yield from awaitable` expression.
    #     # The iterator (generator) yields until exhausted,
    #     # then raises StopIteration with the value returned in by the generator
    #     function.
    #     # return self.result()  # May raise too.
    #     # # Otherwise, the only allowed value from the iterator is None.


_shared_scope_lock = threading.RLock()

_shared_scope_count = contextvars.ContextVar("_shared_scope_count", default=0)

_dispatcher: contextvars.ContextVar = contextvars.ContextVar("_dispatcher")
"""Identify an asynchronous Context.

Non-asyncio-aware functions may need to behave
differently when we know that asynchronous context switching could happen.
We allow multiple dispatchers to be active, but each dispatcher must
1. contextvars.copy_context()
2. set itself as the dispatcher in the new Context.
3. run within the new Context.
4. ensure the Context is destroyed (remove circular references)
"""

current_scope: contextvars.ContextVar = contextvars.ContextVar("current_scope")
"""The active workflow manager, if any.

This property is global within the interpreter or a `contextvars.Context`.

Note that it makes no sense to start a dispatching session without concurrency,
so we can think in terms of a parent context doing contextvars.copy_context().run(...),
but the parent must set the current_scope correctly in the new Context.
"""


def get_scope():
    """Get a reference to the manager of the current workflow scope."""
    # get_scope() is used to determine the default workflow manager when *context*
    # is not provided to scalems object factories, scalems.run(), scalems.wait() and
    # (non-async) `result()` methods.
    # Async coroutines can safely use get_scope(), but should not use the
    # non-async scope() context manager for nested scopes without wrapping
    # in a contextvars.run() that locally resets the Context current_scope.
    try:
        _scope: typing.Union[WorkflowManager, weakref.ReferenceType] = current_scope.get()
        if isinstance(_scope, weakref.ReferenceType):
            _scope = _scope()
        manager: WorkflowManager = _scope
        # This check is in case we use weakref.ref:
        if manager is None:
            raise ProtocolError("Context for current scope seems to have disappeared.")
        logger.debug(f"Scope queried with get_scope() {repr(manager)}")
    except LookupError:
        logger.debug("Scope was queried, but has not yet been set.")
        return None
    return manager


@contextlib.contextmanager
def scope(manager, close_on_exit=False):
    """Set the current workflow management within a clear scope.

    Restores the previous workflow management scope on exiting the context manager.

    To allow better tracking of dispatching chains, this context manager does not allow
    the global workflow management scope to be "stolen". If *manager* is already the
    current scope, a recursion depth is tracked, and the previous scope is restored
    only when the last "scope" context manager for *manager* exits. Multiple "scope"
    context managers are not allowed for different *manager* instances in the same
    :py:class:`context <contextvars.Context>`.

    If *close_on_exit=True*, calls ``manager.close()`` when leaving *manager's*
    scope for the last time, if possible.

    Note:
        Scope indicates the "active" WorkflowManager instance.
        This is separate from WorkflowManager lifetime and ownership.
        WorkflowManagers should track their own activation status and provide logic for
        whether to allow reentrant dispatching.

    Within the context managed by *scope*, get_scope() will return *context*.

    While this context manager should be thread-safe, in general, this context manager
    should only be used in the root thread where the UI and event loop are running to
    ensure we can clean up properly.
    Dispatchers may provide environments in which this context manager can be used in
    non-root threads, but the Dispatcher needs to curate the contextvars.ContextVars
    and ensure that the Context is properly cleaned up.
    """
    _close = None
    if close_on_exit:
        _close = getattr(manager, "close", None)
        if not callable(_close):
            raise ValueError("close_on_exit is True, but manager has no close() method.")

    logger.debug(f"Request to enter the scope of {manager}.")
    with _shared_scope_lock:
        parent = get_scope()
        dispatcher = _dispatcher.get(None)
        # A dispatcher can explicitly allow or disallow nested scopes by setting
        # current_scope to itself or something else.
        if parent is None or parent is dispatcher or parent is manager:
            if parent is not manager:
                logger.debug("Entering scope of {}".format(str(manager)))
            token = current_scope.set(weakref.ref(manager))
            _shared_scope_count.set(_shared_scope_count.get() + 1)
        else:
            assert dispatcher is not None
            raise ProtocolError(f"Cannot nest {manager} scope in {parent} scope while dispatching under {dispatcher}.")
        try:
            yield manager
        finally:
            _shared_scope_count.set(_shared_scope_count.get() - 1)
            if _shared_scope_count.get() == 0:
                logger.debug("Leaving scope of {}".format(str(manager)))
                token.var.reset(token)
                if _close:
                    _close()
    # If we need to have multiple nested scopes across different threads, we can
    # hold the lock only for __enter__ and __exit__, but we will need to get the
    # lock object from a ContextVar or from the dispatcher. Then the above `try`
    # block would be moved up a level and would look something like the following.
    # try:
    #     yield manager
    # finally:
    #     with _shared_scope_lock:
    #         _shared_scope_count.set(_shared_scope_count.get() - 1)
    #         if _shared_scope_count.get() == 0:
    #             logger.debug('Leaving scope of {}'.format(str(manager)))
    #             token.var.reset(token)


def _unpack_work(ref: dict):
    """Temporary handler for ad hoc dict-based input.

    Unpack and serialize the nested task descriptions.

    Note: this assumes work is nested, with only one item per "layer".
    """
    assert isinstance(ref, dict)
    implementation_identifier = ref.get("implementation", None)
    message: dict = ref.get("message", None)
    if not isinstance(implementation_identifier, list) or not isinstance(message, dict):
        raise _exceptions.DispatchError("Bug: bad schema checking?")

    command = implementation_identifier[-1]
    logger.debug(f"Unpacking a {command}")
    # Temporary hack for ad hoc schema.
    if command == "Executable":
        # generate Subprocess
        from scalems.subprocess import SubprocessInput, Subprocess

        input_node, task_node, output_node = message["Executable"]
        kwargs = {
            "argv": input_node["data"]["argv"],
            "stdin": input_node["data"]["stdin"],
            "stdout": task_node["data"]["stdout"],
            "stderr": task_node["data"]["stderr"],
            "environment": input_node["data"]["environment"],
            "resources": task_node["input"]["resources"],
        }
        bound_input = SubprocessInput(**kwargs)
        item = Subprocess(input=bound_input)
        yield item
        return item.uid()
    else:
        # If record bundles dependencies, identify them and yield them first.
        try:
            depends = ref["message"][command]["input"]
        except AttributeError:
            depends = None
        if depends is not None:
            logger.debug(f"Recursively unpacking {depends}")
            dependency: typing.Optional[bytes] = yield from _unpack_work(depends)
        else:
            dependency = None
        if "uid" not in ref:
            ref["uid"] = next_monotonic_integer().to_bytes(32, "big")
        uid: bytes = ref["uid"]
        if dependency is not None:
            logger.debug("Replacing explicit input in {} with reference: {}".format(uid.hex(), dependency.hex()))
            ref["message"][command]["input"] = dependency
        # Then yield the dependent item.
        yield ref
        return uid


@functools.singledispatch
def _wait(ref, *, manager):
    """Use the indicated workflow manager to resolve a reference to a workflow item."""
    raise _exceptions.DispatchError("No dispatcher for this type of reference.")
    # TODO: Return an object supporting the result type interface.


@_wait.register
def _(ref: dict, *, manager):
    # First draft: monolithic implementation directs the workflow manager to add tasks and execute them.
    # TODO: Use a WorkflowManager interface from the core data model.
    if not isinstance(manager, WorkflowManager):
        raise _exceptions.ProtocolError("Provided manager does not implement the required interface.")
    for item in _unpack_work(ref):
        view = manager.add_item(item)
        logger.debug("Added {}: {}".format(view.uid().hex(), str(item)))
    # TODO: If dispatcher is running, wait for the results.
    # TODO: If dispatcher is not running, can we trigger it?


def wait(ref):
    """Resolve a workflow reference to a local object.

    *wait* signals to the SCALE-MS framework that it is time to intervene and
    do some workflow execution management.

    ScaleMS commands return abstract references to work without waiting for the
    work to execute. Other ScaleMS commands can operate on these references,
    relying on the framework to manage data flow.

    If you need to extract a concrete result, or otherwise force data flow resolution
    (blocking the current code until execution and data transfer are complete),
    you may use scalems.wait(ref) to convert a workflow reference to a concrete
    local result.

    Note that scalems.wait() can allow the current scope to yield to other tasks.
    Developers should use scalems.wait() instead of native concurrency primitives
    when coding for dynamic data flow.
    However, the initial implementation does not inspect the context to allow
    such context-sensitive behavior.

    .. todo:: Establish stable API/CPI for tasks that create other tasks or modify the data flow graph
        during execution.

    scalems.wait() will produce an error if you have not configured and launched
    an execution manager in the current scope.

    .. todo:: Acquire asyncio event loop from WorkflowManager.
        scalems.wait is primarily intended as an abstraction from
        https://docs.python.org/3.8/library/asyncio-eventloop.html#asyncio.loop.run_until_complete
        and an alternative to `await`.
    """
    context = get_scope()
    if context is None:
        # Bail out.
        raise _exceptions.DispatchError(str(ref))
    if not isinstance(context, WorkflowManager):
        raise _exceptions.ProtocolError("Expected WorkflowManager. Got {}".format(repr(context)))

    # Dispatch on reference type.
    return _wait(ref, manager=context)
    # TODO: Normalize a Future behavior. Require argument types that guarantee
    #   that the work will eventually be handled, and defer to asyncio.wait, or
    #   reliably produce an error if this is not possible.



"""Test our ability to call a packaged function call through our CLI handler.

Example:
    python -m scalems.call package.json
"""
import logging
import os
import urllib.parse
from subprocess import CompletedProcess
from subprocess import run as subprocess_run
import sys
import tempfile

import scalems.call

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


sample_call_input = dict(
    func=subprocess_run,
    kwargs={"args": ["/bin/echo", "hello", "world"], "capture_output": True},
)


def test_call_builtin_func():
    """Test the API for scalems.call"""
    call = scalems.call.serialize_call(**sample_call_input)
    result = scalems.call.main(scalems.call.deserialize_call(call))
    assert result.exception is None
    complete_process: CompletedProcess = result.return_value
    # Note that we are not treating OS text encoding generally here. However,
    # this is a toy example in which we are just using `subprocess.run` as an
    # arbitrary Python callable for testing. Let's not generalize unless it
    # becomes and issue for devs in other locales.
    assert "hello world" in complete_process.stdout.decode(encoding="utf8")

    def local_echo(*text: str) -> tuple[str]:
        if isinstance(text, str):
            text = (text,)
        if not isinstance(text, tuple) or not all(isinstance(s, str) for s in text):
            raise ValueError("Function call requires one or more positional string arguments.")
        return tuple(text)

    try:
        local_echo(1)
    except ValueError as e:
        reference_error = repr(e)
    else:
        reference_error = None
    assert reference_error is not None

    sample_text_tuple = ("hi", "there")
    call = scalems.call.serialize_call(func=local_echo, args=sample_text_tuple)
    result = scalems.call.main(scalems.call.deserialize_call(call))
    assert result.exception is None
    output = result.return_value
    assert output == sample_text_tuple

    call = scalems.call.serialize_call(func=local_echo, args=(1,))
    result = scalems.call.main(scalems.call.deserialize_call(call))
    assert result.return_value is None
    assert result.exception == reference_error


def test_call_cli(tmp_path):
    """Run the command line in a subprocess to confirm reasonable behavior."""
    call = scalems.call.serialize_call(**sample_call_input)
    outfile = os.path.join(tmp_path, "scalems_out.json")

    with tempfile.NamedTemporaryFile(suffix=".json", mode="w", dir=tmp_path) as tmp_file:
        tmp_file.write(call)
        tmp_file.flush()
        call_path = tmp_file.name
        process = subprocess_run(
            args=(sys.executable, "-m", "scalems.call", call_path, outfile),
            capture_output=True,
            cwd=tmp_path,
        )
    assert process.returncode == 0
    assert os.path.exists(outfile)
    with open(outfile, "r") as fh:
        result: scalems.call.CallResult = scalems.call.deserialize_result(fh.read())
    completed_process: CompletedProcess = result.return_value
    assert "hello world" in completed_process.stdout.decode(encoding="utf8")
    assert tmp_path.samefile(urllib.parse.urlparse(result.directory).path)



"""Execution environment context management."""

__all__ = ("scoped_chdir",)

import contextlib
import logging
import os
import pathlib
import threading
import typing
import warnings

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))

cwd_lock = threading.Lock()


@contextlib.contextmanager
def scoped_chdir(directory: typing.Union[str, bytes, os.PathLike]):
    """Restore original working directory when exiting the context manager.

    Caveats:
        Current working directory is a process-level property. To avoid unexpected
        behavior across threads, only one instance of this context manager may be
        active at a time. If necessary, we could allow for nested cwd contexts,
        but we cannot make this behavior thread-safe.

    See Also:
        https://github.com/SCALE-MS/scale-ms/issues/166
    """
    if isinstance(directory, bytes):
        directory = os.fsdecode(directory)
    path = pathlib.Path(directory)
    if not path.exists() or not path.is_dir():
        raise ValueError(f"Not a valid directory: {str(directory)}")
    if cwd_lock.locked():
        warnings.warn("Another call has already used scoped_chdir. Waiting for lock...")
    with cwd_lock:
        oldpath = os.getcwd()
        os.chdir(path)
        logger.debug(f"Changed current working directory to {path}")
        try:
            yield path
            # If the `with` block using scoped_chdir produces an exception, it will
            # be raised at this point in this function. We want the exception to
            # propagate out of the `with` block, but first we want to restore the
            # original working directory, so we skip `except` but provide a `finally`.
        finally:
            logger.debug(f"Changing working directory back to {oldpath}")
            os.chdir(oldpath)
    logger.info("scoped_chdir exited.")



"""Test the scalems CPI helpers and metaprogramming utilities.

TODO: reconcile and merge with test_datamodel.py
"""

import json
import logging

from scalems.messages import Command, StopCommand

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


def test_serialization():
    cmd = StopCommand()
    serialized = json.dumps(cmd.encode())
    obj: dict = json.loads(serialized)
    assert "control" in obj.keys()
    assert obj["control"] == "stop"


def test_deserialization():
    serialized_stop = '{"control": "stop"}'
    cmd = Command.decode(json.loads(serialized_stop))
    assert cmd.key == "control"
    assert cmd.message == "stop"



"""Workflow subpackage for ScaleMS execution dispatching to RADICAL Pilot.

Manage workflow context for RADICAL Pilot.

Command Line Invocation::

    python -m scalems.radical \\
        --resource local.localhost --venv $HOME/myvenv --access local myworkflow.py

For required and optional command line arguments::

    python -m scalems.radical --help

or refer to the web-based documentation for
`scalems.radical command line`.

The user is largely responsible for establishing appropriate
`RADICAL Cybertools <https://radical-cybertools.github.io/>`__
(RCT) software environment at both the client side
and the execution side. See :ref:`rp setup` in the
:doc:`install` for complete documentation.

See Also:
    Refer to https://github.com/SCALE-MS/scale-ms/issues/141
    for the status of `scalems` support for automatic execution environment bootstrapping.

See Also:
    * `scalems.messages`
    * `scalems.execution`
    * `scalems.radical.runtime`
    * https://github.com/SCALE-MS/scale-ms/issues/90

"""
# TODO: Consider converting to a namespace package to improve modularity of
#  implementation.

__all__ = ["configuration", "parser", "workflow_manager"]

import asyncio
import logging

import scalems.workflow
from ..invocation import make_parser as _make_parser
from .runtime import configuration
from .runtime import parser as _runtime_parser

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))

parser = _make_parser(__package__, parents=[_runtime_parser()])


def workflow_manager(loop: asyncio.AbstractEventLoop, directory=None):
    """Manage a workflow context for RADICAL Pilot work loads.

    The rp.Session is created when the Python Context Manager is "entered",
    so the asyncio event loop must be running before then.

    To help enforce this, we use an async Context Manager, at least in the
    initial implementation. However, the implementation is not thread-safe.
    It is not reentrant, but this is not checked. We probably _do_ want to
    constrain ourselves to zero or one Sessions per environment, but we _do_
    need to support multiple Pilots and task submission scopes (_resource
    requirement groups).
    Further discussion is welcome.

    Warning:
        The importer of this module should be sure to import radical.pilot
        before importing the built-in logging module to avoid spurious warnings.
    """
    from .runtime import executor_factory

    return scalems.workflow.WorkflowManager(loop=loop, executor_factory=executor_factory, directory=directory)



"""Check that the basic Context implementation manages state as expected."""

import json
import logging
import os
import tempfile

import pytest
import scalems.context as _context
import scalems.exceptions
from scalems.store import filestore_generator
from scalems.store import FileStore, FileStoreManager


def test_normal_lifecycle(tmp_path, caplog):
    """Test normal Context data store life cycle / state machine.

    initialize_datastore() should succeed when called in a clean directory or when
    the current process already owns the data store.
    """
    datastore = FileStore(directory=tmp_path)
    assert not datastore.closed
    assert datastore.instance == os.getpid()
    with open(datastore.filepath, "r") as fh:
        assert json.load(fh)["instance"] == os.getpid()
    datastore.close()
    assert datastore.closed
    with open(datastore.filepath, "r") as fh:
        data = json.load(fh)
    assert "instance" in data
    assert data["instance"] is None

    generator = filestore_generator(directory=tmp_path)
    datastore = next(generator)()
    # Disallow multiple FileStores per directory.
    with pytest.raises(scalems.exceptions.ContextError):
        assert tmp_path in FileStore._instances
        assert FileStore._instances[tmp_path] is datastore
        FileStore(directory=tmp_path)
    datastore.close()
    # finalize_datastore() must be called exactly once for a data store that has been
    # opened.
    with pytest.raises(scalems.store.StaleFileStore):
        datastore.close()
    with pytest.raises(StopIteration):
        with pytest.warns(scalems.exceptions.ScaleMSWarning):
            next(generator)
    del generator

    # Confirm some assumptions about implementation details.
    filepath = tmp_path.joinpath(scalems.store._data_subdirectory, scalems.store._metadata_filename)

    manager = FileStoreManager(tmp_path)
    datastore: scalems.store.FileStore = manager.filestore()
    assert datastore.instance == os.getpid()
    assert datastore.filepath == filepath
    with pytest.raises(AttributeError):
        # `log` attribute is not assignable.
        datastore.log = list()
    # TODO: log interface.
    # datastore.log.append('Testing')
    with caplog.at_level(logging.CRITICAL, logger="scalems.store"):
        assert not datastore.closed
        del manager
        assert datastore.closed
    with open(filepath, "r") as fh:
        metadata: dict = json.load(fh)
        # A finalized record should not have an owning *instance*.
        assert metadata["instance"] is None


def test_nonfinalized(tmp_path, caplog):
    """Failure to call finalize_datastore() should have well-defined behavior."""
    with _context.scoped_chdir(tmp_path):
        manager = FileStoreManager()
        datastore = manager.filestore()
        metadata_path = datastore.filepath

        # Fake a bad shutdown.
        with open(metadata_path, "w") as fp:
            json.dump({"instance": 0}, fp)
        # In the future, more sophisticated tests might check that Singleton behavior or scoping
        # protections aren't violated.
        # Right now, we have to manipulate the filesystem with knowledge of the implementation
        # details in order to effectively test.
        with pytest.raises(scalems.exceptions.ContextError):
            FileStoreManager().filestore()
        # We may want to assert constraints on the filesystem state we expect to encounter
        # even when a lock is left unexpectedly, but initially all we know is that a
        # dangling lock may result from an unclean process termination.
        scalems.store._lock._lock_directory()
        with pytest.raises(scalems.exceptions.ContextError):
            FileStoreManager()
        scalems.store._lock._unlock_directory()

        logger = logging.getLogger("scalems.context._datastore")
        logger.debug(
            "Suppressing an error that would normally occur when we fail to "
            "close the FileStore (that we intentionally broke in this test)."
        )
        with caplog.at_level(logging.CRITICAL, logger="scalems.store"):
            del manager


def test_contention(tmp_path, caplog):
    """Test various ways data stores could collide.

    If two processes try to use the same data store, we should be able to
    detect and prevent it.
    """
    with _context.scoped_chdir(tmp_path):
        manager = FileStoreManager()
        datastore = manager.filestore()
        scalems.store._lock._lock_directory()
        with pytest.raises(scalems.store._lock.LockException):
            datastore.close()
        # This failed call to `close` leaves the datastore in an invalid state.
        with pytest.raises(scalems.exceptions.ContextError):
            assert not datastore.closed

        scalems.store._lock._unlock_directory()
        with caplog.at_level(logging.CRITICAL, logger="scalems.store"):
            # The above failed `close` left the manager in an invalid state.
            del manager

        metadata_path = datastore.filepath
        expected_instance = os.getpid()
        unexpected_instance = expected_instance + 1
        with open(metadata_path, "w") as fp:
            json.dump({"instance": unexpected_instance}, fp)
        with pytest.raises(scalems.exceptions.ContextError):
            FileStoreManager()

        scalems.store._lock._lock_directory()
        with pytest.raises(scalems.exceptions.ContextError):
            FileStoreManager()
        scalems.store._lock._unlock_directory()


def test_nesting(tmp_path):
    """Test various ways data stores could collide.

    We should check for old scalems data store versions or broken/invalid directory
    structures.
    """
    # initialization should check for locks in parent directories.
    with scalems.store._lock.scoped_directory_lock(tmp_path):
        with tempfile.TemporaryDirectory(dir=tmp_path) as path:
            with _context.scoped_chdir(path):
                with pytest.raises(scalems.exceptions.ContextError):
                    FileStoreManager(directory=path)

    # initialization should check for unexpected nesting in the work dir.
    manager = FileStoreManager(directory=tmp_path)
    datastore = manager.filestore()
    root_path = datastore.directory
    assert root_path == tmp_path
    datastore_path = datastore.datastore
    with tempfile.TemporaryDirectory(dir=root_path) as path:
        with _context.scoped_chdir(path):
            with pytest.raises(scalems.exceptions.ContextError):
                FileStoreManager()

    # initialization should check for unexpected nesting in the datastore itself.
    with _context.scoped_chdir(datastore_path):
        with pytest.raises(scalems.exceptions.ContextError):
            FileStoreManager()


def test_recovery(tmp_path):
    """A workflow directory should be re-usable if it was shut down cleanly."""

    # Follow the lifecycle of a workflow session.
    manager = FileStoreManager(directory=tmp_path)
    datastore = manager.filestore()
    metadata_path = datastore.filepath
    del manager

    with open(metadata_path, "r") as fh:
        metadata: dict = json.load(fh)
        assert metadata["instance"] is None

    manager = FileStoreManager(directory=tmp_path)
    assert datastore is not manager.filestore()
    assert datastore.closed
    assert not manager.filestore().closed
    with open(metadata_path, "r") as fh:
        metadata: dict = json.load(fh)
        assert metadata["instance"] == os.getpid()



"""Test the scalems.radical.Runtime state object."""
import logging
import typing
import warnings

import pytest
import radical.pilot as rp

from scalems.exceptions import APIError
from scalems.radical.runtime import Runtime

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@pytest.mark.exhaustive
def test_runtime_normal_instance(rp_task_manager, pilot_description):
    """Set the Runtime.pilot from an rp.Pilot instance."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.task_manager")
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.db.database")
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.session")

        session: rp.Session = rp_task_manager.session

        state = Runtime(session=session)

        state.task_manager(rp_task_manager)

        # We only expect one pilot
        pilot: rp.Pilot = rp_task_manager.get_pilots()[0]
        # We get a dictionary...
        # assert isinstance(pilot, rp.Pilot)
        # But it looks like it has the pilot id in it.
        pilot_uid = typing.cast(dict, pilot)["uid"]
        pmgr_uid = typing.cast(dict, pilot)["pmgr"]
        pmgr: rp.PilotManager = session.get_pilot_managers(pmgr_uids=pmgr_uid)
        assert isinstance(pmgr, rp.PilotManager)

        state.pilot_manager(pmgr)

        pilot = pmgr.get_pilots(uids=pilot_uid)
        assert isinstance(pilot, rp.Pilot)
        state.pilot(pilot)


@pytest.mark.exhaustive
def test_runtime_normal_uid(rp_task_manager, pilot_description):
    """Set the Runtime.pilot from the UID obtained from the task_manager."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.task_manager")
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.db.database")
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.session")

        session: rp.Session = rp_task_manager.session

        state = Runtime(session=session)

        state.task_manager(rp_task_manager.uid)

        # We only expect one pilot
        pilot: rp.Pilot = rp_task_manager.get_pilots()[0]
        # We get a dictionary...
        # assert isinstance(pilot, rp.Pilot)
        # But it looks like it has the pilot id in it.
        pilot_uid = typing.cast(dict, pilot)["uid"]

        # It is an error to set a Pilot before the PilotManager has been set.
        with pytest.raises(APIError):
            state.pilot(pilot_uid)

        pmgr_uid = typing.cast(dict, pilot)["pmgr"]

        state.pilot_manager(pmgr_uid)

        state.pilot(pilot_uid)


@pytest.mark.exhaustive
def test_runtime_bad_uid(pilot_description):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.task_manager")
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.db.database")
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.session")

        session = rp.Session()

        with session:
            state = Runtime(session=session)

            with pytest.raises(ValueError):
                state.task_manager("spam")

            tmgr = rp.TaskManager(session=session)
            state.task_manager(tmgr)

            with pytest.raises(ValueError):
                state.pilot_manager("spam")

            pmgr = rp.PilotManager(session=session)
            state.pilot_manager(pmgr)

            with pytest.raises(ValueError):
                state.pilot_manager("spam")

            tmgr.close()
            pmgr.close()

        assert session.closed


@pytest.mark.exhaustive
def test_runtime_mismatch(pilot_description):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.task_manager")
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.db.database")
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="radical.pilot.session")

        session = rp.Session()

        with session:
            original_pmgr = rp.PilotManager(session=session)
            pilot = original_pmgr.submit_pilots(rp.PilotDescription(pilot_description))
            original_tmgr = rp.TaskManager(session=session)
            original_tmgr.add_pilots(pilot)

        assert session.closed
        # This assertion may not be true:
        # assert pilot.state in rp.FINAL
        # Note that Pilot and other components may still be shutting down, but the
        # intention is that, from this point, pmgr, pilot, and tmgr are now "stale".

        session = rp.Session()

        with session:
            state = Runtime(session=session)

            with pytest.raises(APIError):
                state.task_manager(original_tmgr)
            original_tmgr.close()

            tmgr = rp.TaskManager(session=session)
            state.task_manager(tmgr)

            with pytest.raises(APIError):
                state.pilot_manager(original_pmgr)
            original_pmgr.close()

            pmgr = rp.PilotManager(session=session)
            state.pilot_manager(pmgr)

            # The UID will not resolve in the stored PilotManager.
            with pytest.raises(ValueError):
                state.pilot(pilot.uid)

            # The Pilot is detectably invalid.
            with pytest.raises(APIError):
                state.pilot(pilot)

            # Even here, the old Pilot may still be in 'PMGR_ACTIVE_PENDING'
            if pilot.state not in rp.FINAL:
                pilot.cancel()
            tmgr.close()
            pmgr.close()
        assert session.closed
